# The following package is packaged automatically with Flask lambda.
# Do not introduce dependencies outside Python standard library.
from __future__ import annotations
from typing import TYPE_CHECKING
import json
import io
import sys
import base64
from urllib.parse import urlencode

if TYPE_CHECKING:
    from typing import Any


class FlaskLambdaHandler:
    """Flask lambda handler."""

    def __init__(self, app: Any) -> None:
        """Initialize a Flask lambda handler.

        :param app: a Flask app
        """
        self.app = app
        self.status = None
        self.response_headers = None

    def start_response(self, status, response_headers, exc_info=None):
        """Implement Flask callback to store the response.

        See Flask documentation.
        """
        self.status = int(status[:3])
        self.response_headers = dict(response_headers)

    def lambda_handler(self, event, context):
        """Lambda entry point."""
        self.status = None
        self.response_headers = None

        body = next(
            self.app.wsgi_app(
                self.create_flask_wsgi_environ(event, context), self.start_response
            )
        )
        return {
            "statusCode": self.status,
            "headers": self.response_headers,
            "body": body,
        }

    def create_flask_wsgi_environ(self, event: dict, context: dict) -> dict:
        """Create a WSGI environment from AWS lambda input.

        Currently this function supports creation of WSGI environment from
        API Gateway HTTP API 2.0 and a REST API

        :param event: as received by the lambda
        :param context: as received by the lambda
        """
        request_ctx = event["requestContext"]
        remote_user: str | None = None

        # http is True if the event comes from HTTP API gateway
        # otherwise it is false and the event is from a REST API
        http = "version" in event

        if "authorizer" in request_ctx:
            remote_user = request_ctx["authorizer"].get("principalId")
        elif "identity" in request_ctx:
            remote_user = request_ctx["identity"].get("userArn")

        # Compute script_name and path
        path = event["rawPath" if http else "path"]
        script_name = ""
        stage = request_ctx.get("stage", "$default")
        if stage not in ["$default", "default"]:
            script_name = f"/{stage}"
            path = path.replace(script_name, "", 1)

        # HTTP method used
        http_method = (
            request_ctx["http"]["method"] if http else request_ctx["httpMethod"]
        )

        # Normalized headers
        headers = {k.title(): v for k, v in event["headers"].items()}

        # Body
        body = event.get("body", "")
        if event.get("isBase64Encoded", "false") == "true":
            body = base64.b64decode(body)
        elif body:
            body = body.encode("utf-8")
        else:
            body = b""

        query_string_param = event.get("multiValueQueryStringParameters")
        environ = {
            "PATH_INFO": path,
            "QUERY_STRING": event["rawQueryString"]
            if http
            else urlencode(query_string_param, doseq=True)
            if query_string_param
            else "",
            "REMOTE_ADDR": request_ctx["identity"]["sourceIp"],
            "REQUEST_METHOD": http_method,
            "SCRIPT_NAME": script_name,
            "HTTP_HOST": headers["Host"],
            "SERVER_NAME": headers["Host"],
            "SERVER_PORT": headers.get("X-Forwarded-Port", "80"),
            "SERVER_PROTOCOL": str("HTTP/1.1"),
            "wsgi.version": (1, 0),
            "wsgi.url_scheme": headers.get("X-Forwarded-Proto", "http"),
            "wsgi.input": io.BytesIO(body),
            "wsgi.errors": sys.stdout,
            "wsgi.multiprocess": False,
            "wsgi.multithread": False,
            "wsgi.run_once": False,
        }

        # Set content_type and content_length if necessary
        if http_method in ["POST", "PUT", "PATCH", "DELETE"]:
            if "Content-Type" in headers:
                environ["CONTENT_TYPE"] = headers["Content-Type"]
            environ["CONTENT_LENGTH"] = str(len(body))

        # Export headers into the WSGI environment
        for header in headers:
            wsgi_name = "HTTP_" + header.upper().replace("-", "_")
            environ[wsgi_name] = headers[header]

        # Set REMOTE_USER if necessary
        if remote_user:
            environ["REMOTE_USER"] = remote_user

        # For logging purpose
        print(
            json.dumps(
                {
                    k: v
                    for k, v in environ.items()
                    if k not in ("wsgi.input", "wsgi.errors")
                }
            )
        )
        return environ



from __future__ import annotations
import os
from typing import TYPE_CHECKING

from troposphere import cloudformation

from e3.aws import name_to_id
from e3.aws.troposphere import Construct
from e3.aws.troposphere import Stack


if TYPE_CHECKING:  # all: no cover
    from troposphere import AWSHelperFn, AWSObject


class StackSet(Construct):
    """A CloudFormation service managed stack set."""

    def __init__(
        self,
        name: str,
        description: str,
        regions: list[str],
        ous: list[str] | None = None,
        accounts: list[str] | None = None,
        failure_tolerance_count: int = 0,
        max_concurrent_count: int = 1,
    ):
        """Initialize a CloudFormation service managed stack set.

        :param name: stack set name
        :param description: stack set description
        :param regions: list of regions where to deploy stack set stack instances
        :param ous: OrganizationalUnitIds for which to create stack instances
            in the specified Regions. Note that if both ous and accounts parameters
            are None then the stack set is deployed into the whole organisation
        :param accounts: list of accounts where to deploy stack set stack instances
        :param failure_tolerance_count: The number of accounts, per Region, for which
            the stackset deployment operation can fail before AWS CloudFormation stops
            the operation in that region
        :param max_conccurent_count: The maximum number of accounts in which to perform
            stackset deployment operations at one time. It is at most one more than the
            FailureToleranceCount.
        """
        self.name = name
        self.description = description
        self.stack = Stack(stack_name=f"{self.name}-stack", cfn_role_arn="stackset")
        self.template_filename = f"{self.name}-template.yaml"
        self.regions = regions
        self.ous = ous
        self.accounts = accounts

        self.operation_preferences = cloudformation.OperationPreferences(
            FailureToleranceCount=failure_tolerance_count,
            MaxConcurrentCount=max_concurrent_count,
        )

    def add(self, element: AWSObject | Construct | Stack) -> None:
        """Add resource to the stackset stack.

        :param element: resource to add to the stackset stack
        """
        self.stack.add(element)

    def add_condition(self, condition_name: str, condition: AWSHelperFn) -> None:
        """Add condition to stackset stack.

        :param condition_name: name of the condition to add
        :param condition: condition to add
        """
        self.stack.add_condition(condition_name, condition)

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return list of AWSObject associated with the construct."""
        stack_set_args = {
            "AutoDeployment": cloudformation.AutoDeployment(
                Enabled=True, RetainStacksOnAccountRemoval=False
            ),
            "CallAs": "SELF",
            "Capabilities": ["CAPABILITY_NAMED_IAM"],
            "Description": self.description,
            "OperationPreferences": self.operation_preferences,
            "PermissionModel": "SERVICE_MANAGED",
            "StackSetName": self.name,
            "TemplateURL": (
                f"https://{stack.s3_bucket}.s3.amazonaws.com/{stack.s3_key}"
                f"{self.template_filename}"
            ),
        }

        if self.ous is not None or self.accounts is not None:
            stack_instances_group = []

            if self.ous is not None:
                stack_instances_group.append(
                    cloudformation.StackInstances(
                        DeploymentTargets=cloudformation.DeploymentTargets(
                            OrganizationalUnitIds=self.ous
                        ),
                        Regions=self.regions,
                    )
                )

            if self.accounts is not None:
                stack_instances_group.append(
                    cloudformation.StackInstances(
                        DeploymentTargets=cloudformation.DeploymentTargets(
                            Accounts=self.accounts
                        ),
                        Regions=self.regions,
                    )
                )

            stack_set_args["StackInstancesGroup"] = stack_instances_group
        return [cloudformation.StackSet(name_to_id(self.name), **stack_set_args)]

    def create_data_dir(self, root_dir: str) -> None:
        """Create data to be pushed to bucket used by cloudformation for resources."""
        template_path = os.path.join(root_dir, self.template_filename)
        with open(template_path, "w") as template_f:
            template_f.write(self.stack.body)



class CfnConfig(object):
    """A cfn-ini config."""

    def __init__(self):
        """Initialize a config."""
        self.files = {}
        self.commands = []

    def add_file(
        self, filename, mode="644", owner="root", group="root", content=None, path=None
    ):
        """Add a file to a config.

        Either content or path should be set but not both.

        :param filename: path on the target system
        :type filename: str
        :param mode: mode for the file (three digit format like chmod)
        :type mode: str
        :param owner: owner (default: root)
        :type owner: str
        :param group: group (default: root)
        :type group: str
        :param content: file content
        :type content: str | None
        :param path: path to file containing the content
        :type path: str | None
        """
        assert content is None or path is None, "cannot set both path and content"
        if path is not None:
            with open(path, "r") as fd:
                content = fd.read()
        assert content is not None, "no content for file"

        self.files[filename] = {
            "mode": "000" + mode,
            "owner": owner,
            "group": group,
            "content": content,
        }

    def add_s3_file(
        self, filename, url, authentication, mode="644", owner="root", group="root"
    ):
        """Add file to a config where file body is on S3.

        :param filename: path on the target system
        :type filename: str
        :param url: s3 url to the resource. Should be in the format:
            https://[bucket].s3.amazonaws.com/[key]
        :type url: str
        :param authentication: name of an authentication method to use
        :type authentication: str
        :param mode: mode for the file (three digit format like chmod)
        :type mode: str
        :param owner: owner (default: root)
        :type owner: str
        :param group: group (default: root)
        :type group: str
        """
        self.files[filename] = {
            "mode": "000" + mode,
            "owner": owner,
            "group": group,
            "source": url,
            "authentication": authentication,
        }

    def add_command(self, command):
        """Add a command to the config.

        Commands are run in the order in which they are added

        :param command: command to launch
        :type command: str
        """
        self.commands.append(command)

    @property
    def properties(self):
        """Serialize object.

        Used to export to cloudformation.

        :rtype: dict
        """
        result = {"files": self.files, "commands": {}}
        for index, command in enumerate(self.commands):
            # Command name define the order in which they are executed.
            # Alphanumeric order thus the need for leading zeros when
            # generating the command names based on integers.
            result["commands"]["%04d" % index] = {}
            result["commands"]["%04d" % index]["command"] = command
        return result


class CfnIni(object):
    """Represent an AWS::CloudFormation::Init resource."""

    def __init__(self):
        """Initialize cfn-ini data."""
        self.config_sets = {}
        self.configs = {}

    def add_config_set(self, name, config_list):
        """Add a config set.

        :param name: name of the config set
        :type name: str
        :param config_list: list of config names
        :type config_list: list[str]
        """
        for config in config_list:
            assert config in self.configs
        self.config_sets[name] = config_list

    def add_config(self, name, config):
        """Add a configuration.

        :param name: configuration name
        :type name: str
        :param config: config content
        :type config: CfnConfig
        """
        assert isinstance(config, CfnConfig)
        assert name != "configSets", "configSets is not a valid config name"
        self.configs[name] = config

    @property
    def properties(self):
        """Serialize object.

        Used to export to cloudformation.

        :rtype: dict
        """
        result = {}
        if self.config_sets:
            result["configSets"] = self.config_sets
        for name, config in self.configs.items():
            result[name] = config.properties
        return result



"""Provide ECR Repository."""

from __future__ import annotations
from dataclasses import dataclass, field
from typing import TYPE_CHECKING


from troposphere import AWSObject, ecr, Tags

from e3.aws import name_to_id
from e3.aws.troposphere import Construct

if TYPE_CHECKING:
    from e3.aws.troposphere import Stack


@dataclass(frozen=True)
class Repository(Construct):
    """Define a ECR Repository construct.

    :param name: name of the repository
    :param tags: An array of key-value pairs to apply to this resource
    """

    name: str
    tags: dict[str, str] = field(default_factory=lambda: {})

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Construct and return a ECR Repository."""
        return [
            ecr.Repository(
                name_to_id(self.name),
                ImageScanningConfiguration=ecr.ImageScanningConfiguration(
                    ScanOnPush=True
                ),
                ImageTagMutability="IMMUTABLE",
                RepositoryName=self.name,
                Tags=Tags({"Name": self.name, **self.tags}),
            )
        ]



from __future__ import annotations

from typing import TYPE_CHECKING

import json
import mimetypes
import tempfile
from contextlib import closing


from e3.event import EventHandler, unique_id
from e3.fs import rm
import e3.log
from e3.aws import Session

if TYPE_CHECKING:
    from typing import Any
    from e3.event import Event

logger = e3.log.getLogger("S3Handler")


class S3Handler(EventHandler):
    """Event handler that relies on AWS S3."""

    def __init__(
        self,
        event_bucket: str,
        log_bucket: str,
        sse: str = "AES256",
        profile: str | None = None,
    ) -> None:
        """Initialize handler.

        :param event_bucket: event bucket name
        :param log_bucket: log bucket name
        :param sse: encryption method
        :param profile: profile to use for S3 operations
        """
        self.event_bucket = event_bucket
        self.log_bucket = log_bucket
        self.aws_profile = profile
        self.sse = sse

    @classmethod
    def decode_config(cls, config_str: str) -> dict[str, str | None]:
        event_bucket, log_bucket, sse, aws_profile = config_str.split(",", 3)
        return {
            "event_bucket": event_bucket,
            "log_bucket": log_bucket,
            "sse": sse,
            "profile": aws_profile if aws_profile else None,
        }

    def encode_config(self) -> str:
        return "{},{},{},{}".format(
            self.event_bucket,
            self.log_bucket,
            self.sse,
            self.aws_profile if self.aws_profile is not None else "",
        )

    def s3_prefix(self, event: Event) -> str:
        """Additional prefix that depends on the event itself.

        This hook allows a user to add a prefix that depends on the event itself.
        Note that sufixes are still automatically computed so distinct events can
        return the same prefix. The final s3 url used will be
        {log_s3_url}/{s3_prefix}{automatic suffix} for logs and
        {event_s3_url}/{s3_prefix}{automatic suffix} for events metadata.

        :param event: an event
        :return: the prefix
        """
        return ""

    def send_event(self, event: Event) -> bool:
        def s3_cp(
            from_path: str,
            s3_key: str,
            bucket: str,
        ) -> str | None:
            """copy file to S3 bucket.

            :param from_path: File to copy
            :param s3_key: destination in bucket
            :param bucket: name of the bucket
            """
            s_options: dict[str, Any] = {"regions": ["eu-west-1"]}
            if self.aws_profile and self.aws_profile.startswith("arn:aws:iam"):
                # If profile is a role, assume the given role.
                # This is necessary for AWS Lambda as there is no profile file.
                session = Session(**s_options).assume_role(
                    self.aws_profile, "S3Session"
                )
            elif self.aws_profile:
                s_options.update({"profile": self.aws_profile})
                session = Session(**s_options)
            else:
                session = Session(**s_options)

            client = session.to_boto3().client("s3")
            try:
                client.upload_file(
                    Filename=from_path,
                    Bucket=bucket,
                    Key=s3_key,
                    ExtraArgs={"ServerSideEncryption": self.sse},
                )
            except Exception:
                logger.exception(f"Cannot upload file: {s3_key}")
                return None

            return f"s3://{bucket}/{s3_key}"

        # Push attachments to s3 and keep track of their url.
        s3_attachs = {}
        for name, attach in list(event.get_attachments().items()):
            attach_path = attach[0]
            # Push the attachment
            key = f"{self.s3_prefix(event)}{event.uid}/{name}"
            s3_url = s3_cp(attach_path, key, self.log_bucket)

            if s3_url is None:
                return False
            else:
                logger.debug(f"Attachment successfully pushed to {s3_url}")
                ctype, encoding = mimetypes.guess_type(attach_path)
                s3_attachs[name] = {
                    "s3_url": s3_url,
                    "encoding": encoding,
                    "ctype": ctype,
                }

        # Create the JSON to send on the event bucket
        s3_event = {"attachments": s3_attachs, "event": event.as_dict()}

        try:
            tempfile_name = None
            with closing(tempfile.NamedTemporaryFile(mode="w", delete=False)) as fd:
                tempfile_name = fd.name
                json.dump(s3_event, fd)

            # Note that an event can be sent several times with a different
            # status. As a consequence the target url in s3 should be different
            # for call to send.
            key = f"{self.s3_prefix(event)}{event.uid}-{unique_id()}.s3"
            s3_url = s3_cp(tempfile_name, key, self.event_bucket)

            if s3_url is None:
                return False
            else:
                logger.debug(f"Event successfully pushed to {s3_url}")
                return True
        finally:
            if tempfile_name is not None:
                rm(tempfile_name)



"""Provide a VPC construct to run ECS tasks."""
from __future__ import annotations
from dataclasses import dataclass, field
from typing import TYPE_CHECKING


from troposphere import AWSObject, ec2, Ref, Tags

from e3.aws import name_to_id
from e3.aws.troposphere import Construct
from e3.aws.troposphere.iam.policy_document import PolicyDocument
from e3.aws.troposphere.iam.policy_statement import PolicyStatement

if TYPE_CHECKING:
    from e3.aws.troposphere import Stack


@dataclass(frozen=True)
class EcsVPC(Construct):
    """Define a VPC for ECS tasks.

    :param name: name of the VPC
    :param region: region where to deploy the VPC
    :param cidr_block: the primary IPv4 CIDR block for the VPC
    :param subnet_cidr_block: The IPv4 CIDR block assigned to the subnet
    :param tags: the tags for the VPC
    """

    name: str
    region: str
    cidr_block: str = "10.0.0.0/16"
    subnet_cidr_block: str = "10.0.0.0/24"
    tags: dict[str, str] = field(default_factory=lambda: {})

    @property
    def vpc(self) -> ec2.VPC:
        """Return the VPC."""
        return ec2.VPC(
            name_to_id(self.name),
            CidrBlock=self.cidr_block,
            EnableDnsHostnames="true",
            EnableDnsSupport="true",
            Tags=Tags({"Name": self.name, **self.tags}),
        )

    @property
    def subnet(self) -> ec2.Subnet:
        """Return a Subnet for the VPC."""
        return ec2.Subnet(
            name_to_id(f"{self.name}Subnet"),
            VpcId=Ref(self.vpc),
            CidrBlock=self.subnet_cidr_block,
        )

    # Security group and traffic control
    @property
    def security_group(self) -> ec2.SecurityGroup:
        """Return a security group for ECS tasks."""
        return ec2.SecurityGroup(
            name_to_id(f"{self.name}SecurityGroup"),
            GroupDescription="Security group for ECS tasks",
            SecurityGroupEgress=[],
            SecurityGroupIngress=[],
            VpcId=Ref(self.vpc),
        )

    @property
    def ingress_rule(self) -> ec2.SecurityGroupIngress:
        """Return Ingress rule allowing traffic from aws VPC endpoints."""
        return ec2.SecurityGroupIngress(
            name_to_id(f"{self.name}Ingress"),
            CidrIp=self.cidr_block,
            FromPort="443",
            ToPort="443",
            IpProtocol="tcp",
            GroupId=Ref(self.security_group),
        )

    @property
    def egress_rule(self) -> ec2.SecurityGroupEgress:
        """Return egress that disables default egress Rule."""
        return ec2.SecurityGroupEgress(
            name_to_id(f"{self.name}Egress"),
            CidrIp=self.cidr_block,
            IpProtocol="-1",
            GroupId=Ref(self.security_group),
        )

    @property
    def s3_egress_rule(self) -> ec2.SecurityGroupEgress:
        """Return Egress rule allowing traffic to s3."""
        return ec2.SecurityGroupEgress(
            name_to_id(f"{self.name}S3Egress"),
            DestinationPrefixListId="pl-6da54004",
            FromPort="443",
            ToPort="443",
            IpProtocol="tcp",
            GroupId=Ref(self.security_group),
        )

    # VPC Enpoints
    @property
    def s3_route_table(self) -> ec2.RouteTable:
        """Return a route table for s3 endpoint."""
        return ec2.RouteTable(
            name_to_id(f"{self.name}S3RouteTable"), VpcId=Ref(self.vpc)
        )

    @property
    def s3_route_table_assoc(self) -> ec2.SubnetRouteTableAssociation:
        """Return route table association."""
        return ec2.SubnetRouteTableAssociation(
            name_to_id(f"{self.name}S3RouteTableAssoc"),
            RouteTableId=Ref(self.s3_route_table),
            SubnetId=Ref(self.subnet),
        )

    @property
    def s3_vpc_endpoint(self) -> ec2.VPCEndPoint:
        """Return S3 VPC Endpoint.

        Needed to retrieve images as ECR store images on S3.
        """
        return ec2.VPCEndpoint(
            name_to_id(f"{self.name}S3Endpoint"),
            PolicyDocument=self.s3_endpoint_policy_document.as_dict,
            RouteTableIds=[Ref(self.s3_route_table)],
            ServiceName=f"com.amazonaws.{self.region}.s3",
            VpcEndpointType="Gateway",
            VpcId=Ref(self.vpc),
        )

    @property
    def s3_endpoint_policy_document(self) -> PolicyDocument:
        """Return policy document for S3 endpoint."""
        return PolicyDocument(
            statements=[
                PolicyStatement(
                    action=["s3:GetObject", "s3:ListBucket"],
                    effect="Allow",
                    resource="*",
                    principal="*",
                )
            ]
        )

    @property
    def sts_vpc_endpoint(self) -> ec2.VPCEndPoint:
        """Return STS VPC Endpoint.

        Needed to initialize session from the container.
        """
        return ec2.VPCEndpoint(
            name_to_id(f"{self.name}STSEndpoint"),
            PrivateDnsEnabled="true",
            SecurityGroupIds=[Ref(self.security_group)],
            ServiceName=f"com.amazonaws.{self.region}.sts",
            SubnetIds=[Ref(self.subnet)],
            VpcEndpointType="Interface",
            VpcId=Ref(self.vpc),
        )

    @property
    def cloudwatch_logs_vpc_endpoint(self) -> ec2.VPCEndPoint:
        """Return Cloudwatch VPC Endpoint.

        Needed for ecs task to send logs to cloudwatch.
        """
        return ec2.VPCEndpoint(
            name_to_id(f"{self.name}CloudwatchLogsEndpoint"),
            PrivateDnsEnabled="true",
            SecurityGroupIds=[Ref(self.security_group)],
            ServiceName=f"com.amazonaws.{self.region}.logs",
            SubnetIds=[Ref(self.subnet)],
            VpcEndpointType="Interface",
            VpcId=Ref(self.vpc),
        )

    @property
    def ecr_dkr_vpc_endpoint(self) -> ec2.VPCEndPoint:
        """Return ECR dkr VPC Endpoint.

        Needed to retrieve images from ECR service.
        """
        return ec2.VPCEndpoint(
            name_to_id(f"{self.name}EcrDkrEndpoint"),
            PolicyDocument=self.ecr_endpoints_policy_document.as_dict,
            PrivateDnsEnabled="true",
            SecurityGroupIds=[Ref(self.security_group)],
            ServiceName=f"com.amazonaws.{self.region}.ecr.dkr",
            SubnetIds=[Ref(self.subnet)],
            VpcEndpointType="Interface",
            VpcId=Ref(self.vpc),
        )

    @property
    def ecr_api_vpc_endpoint(self) -> ec2.VPCEndPoint:
        """Return ECR api VPC Endpoint.

        Needed to retrieve images from ECR service.
        """
        return ec2.VPCEndpoint(
            name_to_id(f"{self.name}EcrApiEndpoint"),
            PolicyDocument=self.ecr_endpoints_policy_document.as_dict,
            PrivateDnsEnabled="true",
            SecurityGroupIds=[Ref(self.security_group)],
            ServiceName=f"com.amazonaws.{self.region}.ecr.api",
            SubnetIds=[Ref(self.subnet)],
            VpcEndpointType="Interface",
            VpcId=Ref(self.vpc),
        )

    @property
    def ecr_endpoints_policy_document(self) -> PolicyDocument:
        """Return policy Document for ecr endpoint allowing only image pulls."""
        return PolicyDocument(
            statements=[
                PolicyStatement(
                    action=[
                        "ecr:BatchGetImage",
                        "ecr:GetAuthorizationToken",
                        "ecr:GetDownloadUrlForLayer",
                    ],
                    effect="Allow",
                    resource="*",
                    principal="*",
                )
            ]
        )

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Construct and return EcsVPC resources."""
        return [
            self.vpc,
            self.subnet,
            self.security_group,
            self.ingress_rule,
            self.egress_rule,
            self.s3_egress_rule,
            self.s3_route_table,
            self.s3_route_table_assoc,
            self.s3_vpc_endpoint,
            self.cloudwatch_logs_vpc_endpoint,
            self.sts_vpc_endpoint,
            self.ecr_dkr_vpc_endpoint,
            self.ecr_api_vpc_endpoint,
        ]



from e3.aws.cfn.ec2 import VPC
from e3.aws.cfn.route53 import HostedZone, RecordSet


def test_create_recordset():
    """RecordSet test."""
    r = RecordSet(
        "myrecordset",
        hosted_zone="example.com.",
        dns_name="myserver.example.com",
        dns_type="A",
        ttl=60,
        resource_records=["1.1.1.1"],
    )
    assert r.properties


def test_create_hosted_zone():
    """Hosted Zone test."""
    public_hz = HostedZone("myhostedzone", "example.com.")
    assert public_hz.properties

    vpc = VPC("myvpc", "192.168.0.0/16")
    private_hz = HostedZone("privhostedzone", "local.com.", vpcs=[vpc])
    assert private_hz.properties



from __future__ import annotations
from typing import TYPE_CHECKING
from troposphere import dynamodb, GetAtt, Ref, Tags
from troposphere.dynamodb import PointInTimeRecoverySpecification

from e3.aws import name_to_id
from e3.aws.troposphere import Construct

if TYPE_CHECKING:
    from troposphere import AWSObject
    from e3.aws.troposphere import Stack


class Table(Construct):
    """A DynamoDB Table."""

    def __init__(
        self,
        name: str,
        attribute_definitions: dict[str, str],
        key_schema: dict[str, str],
        tags: dict[str, str] | None = None,
        point_in_time_recovery_enabled: bool | None = True,
        billing_mode: str | None = None,
        read_capacity_units: int | Ref | None = None,
        write_capacity_units: int | Ref | None = None,
        time_to_live_attribute_name: str | None = None,
        time_to_live_enabled: bool | None = None,
        stream_enabled: bool | None = None,
        stream_view_type: str | None = None,
    ):
        """Initialize an AWS DynamoDB table.

        :param name: table name
        :param attribute_definitions: dictionary for attribute definitions, keys are
            attributes names and values are DynamoDB attributes types
        :param key_schema: specifies the attributes that make up the primary key for
            the table
        :param tags: dictionary of tags to apply to this resource
        :param point_in_time_recovery_enabled: indicates whether point in time
            recovery is enabled (true) or disabled (false) on the table (default True)
        :param billing_mode: specify how you are charged for read and write throughput
            and how you manage capacity
        :param read_capacity_units: the maximum number of strongly consistent reads
            consumed per second before DynamoDB returns a ThrottlingException
            (default 10)
        :param write_capacity_units: the maximum number of writes consumed per second
            before DynamoDB returns a ThrottlingException (default 10)
        :param time_to_live_attribute_name: the name of the TTL attribute used to store
            the expiration time for items in the table
        :param time_to_live_enabled: indicates whether TTL is to be enabled (true) or
            disabled (false) on the table
        :param stream_enabled: indicates whether DynamoDB Streams is enabled (true)
            or disabled (false) on the table
        :param stream_view_type: when an item in the table is modified, StreamViewType
            determines what information is written to the stream for this table
            (default NEW_IMAGE)
        """
        self.name = name
        self.attribute_definitions = attribute_definitions
        self.key_schema = key_schema
        self.tags = tags
        self.point_in_time_recovery_enabled = point_in_time_recovery_enabled
        self.billing_mode = billing_mode
        self.read_capacity_units = read_capacity_units
        self.write_capacity_units = write_capacity_units
        self.time_to_live_attribute_name = time_to_live_attribute_name
        self.time_to_live_enabled = time_to_live_enabled
        self.stream_enabled = stream_enabled
        self.stream_view_type = stream_view_type

    @property
    def arn(self) -> GetAtt:
        """Arn of the DynamoDB table."""
        return GetAtt(name_to_id(self.name), "Arn")

    @property
    def ref(self) -> Ref:
        return Ref(name_to_id(self.name))

    @property
    def stream_arn(self) -> GetAtt:
        """StreamArn of the DynamoDB table."""
        return GetAtt(name_to_id(self.name), "StreamArn")

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return list of AWSObject associated with the construct."""
        params = {
            "TableName": self.name,
            "AttributeDefinitions": [
                dynamodb.AttributeDefinition(AttributeName=k, AttributeType=v)
                for k, v in self.attribute_definitions.items()
            ],
            "KeySchema": [
                dynamodb.KeySchema(AttributeName=k, KeyType=v)
                for k, v in self.key_schema.items()
            ],
        }

        if self.tags is not None:
            params["Tags"] = Tags(**self.tags)

        if self.point_in_time_recovery_enabled:
            params[
                "PointInTimeRecoverySpecification"
            ] = PointInTimeRecoverySpecification(PointInTimeRecoveryEnabled=True)

        if self.billing_mode is not None:
            params["BillingMode"] = self.billing_mode

        params["ProvisionedThroughput"] = dynamodb.ProvisionedThroughput(
            ReadCapacityUnits=self.read_capacity_units
            if self.read_capacity_units is not None
            else 10,
            WriteCapacityUnits=self.write_capacity_units
            if self.write_capacity_units is not None
            else 10,
        )

        if self.time_to_live_enabled is not None:
            assert (
                self.time_to_live_attribute_name is not None
            ), "time_to_live_attribute_name should be set"
            params["TimeToLiveSpecification"] = dynamodb.TimeToLiveSpecification(
                AttributeName=self.time_to_live_attribute_name,
                Enabled=self.time_to_live_enabled,
            )

        if self.stream_enabled:
            params["StreamSpecification"] = dynamodb.StreamSpecification(
                StreamViewType=self.stream_view_type
                if self.stream_view_type is not None
                else "NEW_IMAGE"
            )

        return [dynamodb.Table(name_to_id(self.name), **params)]



"""Provide IAM Role classes."""
from __future__ import annotations
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from troposphere import AWSObject, iam, Tags, GetAtt

from e3.aws import name_to_id
from e3.aws.troposphere import Construct
from e3.aws.troposphere.iam.policy_document import PolicyDocument
from e3.aws.troposphere.iam.policy_statement import AssumeRole, Trust, Allow

if TYPE_CHECKING:
    from e3.aws.troposphere import Stack
    from e3.aws.troposphere.iam.policy_statement import PrincipalType


@dataclass
class Role(Construct):
    """Define IAM role with an attached assume_role policy document.

    :param name: role name
    :param description: role description
    :param trust: trust policy. It can be either a principal, a Trust statement or
        a policy document
    :param managed_policy_arns: list of ARNs of IAM managed policies to attach
        to the role
    :param max_session_duration: the maximum session duration (in seconds) that
        you want to set for the specified role. default is one hour
    :param tags: a list of tags that are attached to the specified role
    :param path: The path to the role.
    :param boundary: The ARN of the policy used to set the permissions boundary
        for the role.
    :param condition: condition contains statements that define the circumstances
        under which role the is created.
    """

    name: str
    description: str
    trust: PrincipalType | Trust | PolicyDocument
    managed_policy_arns: list[str] | None = None
    max_session_duration: int | None = None
    tags: dict[str, str] = field(default_factory=lambda: {})
    path: str = "/"
    boundary: str | None = None
    condition: dict[str, dict] | None = None

    @property
    def trust_policy(self) -> PolicyDocument:
        """Return the trust policy for the role."""
        if isinstance(self.trust, Trust):
            return PolicyDocument(statements=[self.trust])
        elif isinstance(self.trust, PolicyDocument):
            return self.trust
        else:
            return PolicyDocument(statements=[AssumeRole(principal=self.trust)])

    @property
    def arn(self):
        return GetAtt(name_to_id(self.name), "Arn")

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return troposphere objects defining the role."""
        attr = {}

        for key, val in {
            "RoleName": self.name,
            "Description": self.description,
            "ManagedPolicyArns": self.managed_policy_arns,
            "MaxSessionDuration": self.max_session_duration,
            "AssumeRolePolicyDocument": self.trust_policy.as_dict,
            "Tags": Tags({"Name": self.name, **self.tags}),
            "Path": self.path,
            "PermissionsBoundary": self.boundary,
            "Condition": self.condition,
        }.items():
            if val is not None:
                attr[key] = val

        return [iam.Role(name_to_id(self.name), **attr)]

    def cfn_policy_document(self, stack: Stack) -> PolicyDocument:
        return PolicyDocument(
            statements=[
                Allow(
                    action=[
                        "iam:GetRole",
                        "iam:CreateRole",
                        "iam:AttachRolePolicy",
                        "iam:DetachRolePolicy",
                        "iam:DeleteRole",
                    ],
                    resource=f"arn:aws:iam::%(account)s:role/{self.name}",
                )
            ]
        )



import pytest
from e3.aws.cfn.iam import (
    Allow,
    Deny,
    Group,
    InstanceRole,
    Policy,
    PolicyDocument,
    Principal,
    PrincipalKind,
    User,
)
from e3.aws.cfn.s3 import Bucket


def test_create_statements():
    """Various statement creations."""
    s1 = Deny(
        apply_to=Principal(PrincipalKind.SERVICE, "myservice"),
        to="do_something",
        not_on="resource",
    )
    assert s1.properties

    s2 = Allow(
        apply_to=Principal(PrincipalKind.SERVICE, "myservice"),
        sid="id1",
        to="do_something",
        on="allowed_resource",
    )
    assert s2.properties

    pd1 = PolicyDocument()
    pd2 = PolicyDocument()
    pd1.append(s1)
    pd2.append(s2)
    pd3 = pd1 + pd2
    assert len(pd3.statements) == 2
    pd1 += [s2]
    assert len(pd3.statements) == 2


def test_create_instance_profile():
    """Create a basic instance role that get access to a bucket."""
    s = Bucket("MyBucket")
    policy_document = PolicyDocument()
    policy_document.append(
        Allow().to(["s3:ListBucket", "s3:GetObject", "s3:ListObjects"]).on(s.arn)
    )

    instance_profile = InstanceRole("InstRole")
    instance_profile.add_policy(Policy("Pol", policy_document))
    assert instance_profile.body


def test_principal_star():
    """Create a list of principal with one principal being '*'."""
    pl = [
        Principal(PrincipalKind.SERVICE, "ec2.amazonaws.com"),
        Principal(PrincipalKind.EVERYONE),
    ]
    with pytest.raises(AssertionError):
        Principal.property_list(pl)

    pl = [Principal(PrincipalKind.EVERYONE)]
    assert Principal.property_list(pl)


def test_create_user_and_group():
    """Create a basic group."""
    mygroup = Group("mygroup")
    myuser = User("myuser", groups=[mygroup.name])
    assert myuser.properties["Groups"] == [mygroup.name]
    assert mygroup.properties["GroupName"] == mygroup.name



from e3.aws.cfn import AWSType, Resource
from e3.aws.cfn.ec2 import VPC


class PrivateDnsNamespace(Resource):
    def __init__(self, name, vpc, domain, description=None):
        """Initialize a private DNS namespace.

        :param name: logical name in stack
        :type name: str
        :param vpc: vpc in which the namespace is used
        :type vpc: e3.aws.ec2.VPC
        :param domain: domain
        :type domain: str
        :param description: optional description
        :type description: str | None
        """
        super(PrivateDnsNamespace, self).__init__(
            name, kind=AWSType.SERVICE_DISCOVERY_PRIVATE_DNS_NAMESPACE
        )
        self.domain = domain
        assert isinstance(vpc, VPC)
        self.vpc = vpc
        self.description = description

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        result = {"Vpc": self.vpc.ref, "Name": self.domain}
        if self.description:
            result["Description"] = self.description
        return result



from __future__ import absolute_import, division, print_function

import logging
import tempfile

import e3.log
import pytest
from e3.env import Env
from e3.fs import rm
from e3.os.fs import cd

# Activate full debug logs
e3.log.activate(level=logging.DEBUG, e3_debug=True)


@pytest.fixture(autouse=True)
def env_protect(request):
    """Protection against environment change.

    The fixture is enabled for all tests and does the following:

    * store/restore env between each tests
    * create a temporary directory and do a cd to it before each
      test. The directory is automatically removed when test ends
    """
    Env().store()
    tempd = tempfile.mkdtemp()
    cd(tempd)

    def restore_env():
        Env().restore()
        rm(tempd, True)

    request.addfinalizer(restore_env)



"""Provide ecr construct tests."""
import json
import os

from troposphere import ec2, Ref

from e3.aws.troposphere import Stack
from e3.aws.troposphere.ec2 import VPC
from e3.aws.troposphere.iam.policy_statement import Allow
from e3.aws.troposphere.iam.policy_document import PolicyDocument


TEST_DIR = os.path.dirname(os.path.abspath(__file__))


def test_vpc(stack: Stack) -> None:
    """Test VPC creation."""
    ecr_endpoint_pd = PolicyDocument(
        statements=[
            Allow(
                action=[
                    "ecr:BatchGetImage",
                    "ecr:GetAuthorizationToken",
                    "ecr:GetDownloadUrlForLayer",
                ],
                resource="*",
                principal="*",
            )
        ]
    )
    s3_endpoint_pd = PolicyDocument(
        statements=[
            Allow(action=["s3:PutObject", "s3:GetObject"], resource="*", principal="*"),
            Allow(action="s3:ListBucket", resource="*", principal="*"),
        ]
    )
    cloudwatch_endpoint_pd = PolicyDocument(
        statements=[
            Allow(
                action=[
                    "logs:CreateLogStream",
                    "logs:CreateLogGroup",
                    "logs:PutLogEvents",
                ],
                resource="*",
                principal="*",
            )
        ]
    )
    sm_endpoint_pd = PolicyDocument(
        statements=[
            Allow(
                action=[
                    "secretsmanager:GetResourcePolicy",
                    "secretsmanager:GetSecretValue",
                    "secretsmanager:DescribeSecret",
                    "secretsmanager:ListSecretVersionIds",
                ],
                resource=["this_is_a_secret_arn"],
                principal="*",
            )
        ]
    )
    vpc = VPC(
        name="TestVPC",
        region="eu-west-1",
        nat_gateway=True,
        s3_endpoint_policy_document=s3_endpoint_pd,
        interface_endpoints=[
            ("logs", cloudwatch_endpoint_pd),
            ("ecr.api", ecr_endpoint_pd),
            ("ecr.dkr", ecr_endpoint_pd),
            ("sts", None),
            ("secretsmanager", sm_endpoint_pd),
        ],
    )
    stack.add(vpc)

    # Add a security group with access to VPC endpoints
    group_name = "SGWithVPCEndpointsAccess"
    sg = ec2.SecurityGroup(
        group_name,
        GroupDescription="Security group for some privileged runners that need "
        "outbound to the world",
        GroupName=group_name,
        SecurityGroupEgress=vpc.egress_to_vpc_endpoints,
        VpcId=Ref(vpc.vpc),
    )
    stack.add(sg)

    with open(os.path.join(TEST_DIR, "vpc.json")) as fd:
        expected_template = json.load(fd)

    assert stack.export()["Resources"] == expected_template


def test_vpc_with_ses_endpoint(stack: Stack) -> None:
    """Test creation of a VPC with an SES endpoint."""
    vpc = VPC(
        name="TestVPC",
        region="eu-west-1",
        nat_gateway=False,
        interface_endpoints=[
            ("email-smtp", None),
        ],
    )
    stack.add(vpc)

    with open(os.path.join(TEST_DIR, "vpc_ses_endpoint.json")) as fd:
        expected_template = json.load(fd)

    assert stack.export()["Resources"] == expected_template


def test_vpc_with_ses_and_other_endpoints(stack: Stack) -> None:
    """Test creation of a VPC with an SES endpoint and other endpoints."""
    vpc = VPC(
        name="TestVPC",
        region="eu-west-1",
        nat_gateway=False,
        interface_endpoints=[
            ("email-smtp", None),
            ("logs", None),
            ("sts", None),
        ],
    )
    stack.add(vpc)

    with open(os.path.join(TEST_DIR, "vpc_ses_and_other_endpoints.json")) as fd:
        expected_template = json.load(fd)

    assert stack.export()["Resources"] == expected_template



"""Provide S3 buckets."""

from __future__ import annotations
from enum import Enum
from typing import TYPE_CHECKING


from troposphere import AccountId, AWSObject, s3, Ref, GetAtt

from e3.aws import name_to_id
from e3.aws.troposphere.awslambda import Function
from e3.aws.troposphere import Construct
from e3.aws.troposphere.iam.policy_document import PolicyDocument
from e3.aws.troposphere.iam.policy_statement import PolicyStatement
from e3.aws.troposphere.sns import Topic
from e3.aws.troposphere.sqs import Queue


if TYPE_CHECKING:
    from e3.aws.troposphere import Stack
    from e3.aws.troposphere.iam.policy_statement import ConditionType


class EncryptionAlgorithm(Enum):
    """Provide an Enum to describe encryption algorithms."""

    AES256 = "AES256"
    KMS = "aws:kms"


class Bucket(Construct):
    """Define a S3 bucket construct with security parameters and a security policy."""

    def __init__(
        self,
        name: str,
        enable_versioning: bool = True,
        lifecycle_rules: list[s3.LifecycleRule] | None = None,
        default_bucket_encryption: EncryptionAlgorithm
        | None = EncryptionAlgorithm.AES256,
        authorized_encryptions: list[EncryptionAlgorithm] | None = None,
    ):
        """Initialize a bucket.

        :param name: bucket name
        :param enable_versioning: can be set to enable multiple versions of all
            objects in the bucket.
        :param lifecycle_rules: lifecycle rules for bucket objects
        :param default_bucket_encryption: type of the default bucket encryption.
        :param authorized_encryptions: types of the server side encryptions
            to authorize.
        """
        self.name = name
        self.enable_versioning = enable_versioning
        self.lifecycle_rules = lifecycle_rules
        self.default_bucket_encryption = default_bucket_encryption
        if authorized_encryptions is None:
            self.authorized_encryptions = [EncryptionAlgorithm.AES256]
        else:
            self.authorized_encryptions = authorized_encryptions
        self.lambda_configurations: list[
            tuple[dict[str, str], Function | None, str]
        ] = []
        self.topic_configurations: list[tuple[dict[str, str], Topic | None, str]] = []
        self.queue_configurations: list[tuple[dict[str, str], Queue | None, str]] = []
        self.depends_on: list[str] = []

        # Add minimal policy statements
        self.policy_statements = [
            # Deny any request not using https transport protocol
            PolicyStatement(
                action="s3:*",
                effect="Deny",
                resource=self.all_objects_arn,
                principal={"AWS": "*"},
                condition={"Bool": {"aws:SecureTransport": "false"}},
            )
        ]

        assert (
            self.authorized_encryptions
        ), "At least one authorized s3 encryption should be provided"

        # The one element case is needed for retrocompatibility
        # with stacks deployed with older versions of e3-aws
        condition: ConditionType
        if len(self.authorized_encryptions) == 1:
            condition = {
                "StringNotEquals": {
                    "s3:x-amz-server-side-encryption": self.authorized_encryptions[
                        0
                    ].value
                }
            }
        else:
            condition = {
                "ForAllValues:StringNotEquals": {
                    "s3:x-amz-server-side-encryption": [
                        enc.value for enc in self.authorized_encryptions
                    ]
                }
            }

        self.policy_statements.extend(
            [
                # Deny to store object not encrypted with AES256 encryption
                PolicyStatement(
                    action="s3:PutObject",
                    effect="Deny",
                    resource=self.all_objects_arn,
                    principal={"AWS": "*"},
                    condition=condition,
                ),
                # Deny to store non encrypted objects
                # (??? do we really need that statement)
                PolicyStatement(
                    action="s3:PutObject",
                    effect="Deny",
                    resource=self.all_objects_arn,
                    principal={"AWS": "*"},
                    condition={"Null": {"s3:x-amz-server-side-encryption": "true"}},
                ),
            ]
        )

    @property
    def policy_document(self) -> PolicyDocument:
        """Return PolicyDocument to be attached to the bucket."""
        return PolicyDocument(statements=self.policy_statements)

    def add_notification_configuration(
        self,
        event: str,
        target: Function | Topic | Queue | str,
        permission_suffix: str,
        s3_filter: s3.Filter | None = None,
    ) -> None:
        """Add a configuration to bucket notification rules.

        :param event: the S3 bucket event for which to invoke the Lambda function
        :param function: function to invoke when the specified event type occurs
        :param permission_suffix: a name suffix for permissions or policy objects
        :param s3_filter: the filtering rules that determine which objects invoke
            the AWS Lambda function
        """
        params = {"Event": event}
        if s3_filter:
            params["Filter"] = s3_filter

        if isinstance(target, Topic):
            params["Topic"] = target.arn
            self.topic_configurations.append((params, target, permission_suffix))
        elif isinstance(target, Function):
            params["Function"] = target.arn
            self.lambda_configurations.append((params, target, permission_suffix))
        elif isinstance(target, Queue):
            params["Queue"] = target.arn
            self.queue_configurations.append((params, target, permission_suffix))
        elif ":sns:" in target:
            params["Topic"] = target
            self.topic_configurations.append((params, None, permission_suffix))
        elif ":lambda:" in target:
            params["Function"] = target
            self.lambda_configurations.append((params, None, permission_suffix))
        elif ":sqs:" in target:
            params["Queue"] = target
            self.queue_configurations.append((params, None, permission_suffix))

    @property
    def notification_setup(
        self,
    ) -> tuple[s3.NotificationConfiguration, list[AWSObject]]:
        """Return notifcation configuration and associated resources."""
        notification_resources = []
        notification_config = None
        params = {}
        if self.lambda_configurations:
            params.update(
                {
                    "LambdaConfigurations": [
                        s3.LambdaConfigurations(**lambda_params)
                        for lambda_params, _, _ in self.lambda_configurations
                    ]
                }
            )
            # Add Permission invoke for lambdas
            for _, function, suffix in self.lambda_configurations:
                if function:
                    notification_resources.append(
                        function.invoke_permission(
                            name_suffix=suffix,
                            service="s3",
                            source_arn=self.arn,
                            source_account=AccountId,
                        )
                    )
        if self.topic_configurations:
            params.update(
                {
                    "TopicConfigurations": [
                        s3.TopicConfigurations(**topic_params)
                        for topic_params, _, _ in self.topic_configurations
                    ]
                }
            )
            # Add policy allowing to publish to topics
            for _, topic, suffix in self.topic_configurations:
                if topic:
                    topic_policy = topic.allow_publish_policy(
                        service="s3",
                        name_suffix=suffix,
                        condition={"ArnLike": {"aws:SourceArn": self.arn}},
                    )
                    notification_resources.append(topic_policy)
                    self.depends_on.append(topic_policy)
        if self.queue_configurations:
            params.update(
                {
                    "QueueConfigurations": [
                        s3.QueueConfigurations(**queue_params)
                        for queue_params, _, _ in self.queue_configurations
                    ]
                }
            )
            for _, queue, suffix in self.queue_configurations:
                if queue:
                    queue_policy = queue.allow_service_to_write(
                        service="s3",
                        name_suffix=suffix,
                        condition={"ArnLike": {"aws:SourceArn": self.arn}},
                    )
                    notification_resources.append(queue_policy)
                    self.depends_on.append(queue_policy)

        if params:
            notification_config = s3.NotificationConfiguration(
                name_to_id(self.name + "NotifConfig"), **params
            )

        return notification_config, notification_resources

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Construct and return a s3.Bucket and its associated s3.BucketPolicy."""
        # Handle versioning configuration
        optional_resources = []
        versioning_status = "Suspended"
        if self.enable_versioning:
            versioning_status = "Enabled"

        # Block all public accesses
        public_access_block_config = s3.PublicAccessBlockConfiguration(
            BlockPublicAcls=True,
            BlockPublicPolicy=True,
            IgnorePublicAcls=True,
            RestrictPublicBuckets=True,
        )

        # Set default bucket encryption to AES256
        bucket_encryption = None
        if self.default_bucket_encryption:
            bucket_encryption = s3.BucketEncryption(
                ServerSideEncryptionConfiguration=[
                    s3.ServerSideEncryptionRule(
                        ServerSideEncryptionByDefault=s3.ServerSideEncryptionByDefault(
                            SSEAlgorithm=self.default_bucket_encryption.value
                        )
                    )
                ]
            )

        lifecycle_config = None
        if self.lifecycle_rules:
            lifecycle_config = s3.LifecycleConfiguration(
                name_to_id(self.name) + "LifeCycleConfig", Rules=self.lifecycle_rules
            )

        notification_config, notification_resources = self.notification_setup
        optional_resources.extend(notification_resources)

        attr = {}
        for key, val in {
            "BucketName": self.name,
            "BucketEncryption": bucket_encryption,
            "PublicAccessBlockConfiguration": public_access_block_config,
            "VersioningConfiguration": s3.VersioningConfiguration(
                Status=versioning_status
            ),
            "LifecycleConfiguration": lifecycle_config,
            "NotificationConfiguration": notification_config,
            "DependsOn": self.depends_on,
        }.items():
            if val:
                attr[key] = val

        return [
            s3.Bucket(name_to_id(self.name), **attr),
            s3.BucketPolicy(
                name_to_id(self.name) + "Policy",
                Bucket=self.ref,
                PolicyDocument=self.policy_document.as_dict,
            ),
            *optional_resources,
        ]

    @property
    def ref(self):
        return Ref(name_to_id(self.name))

    @property
    def arn(self):
        return f"arn:aws:s3:::{self.name}"

    @property
    def all_objects_arn(self):
        return f"{self.arn}/*"

    @property
    def domain_name(self):
        return GetAtt(name_to_id(self.name), "DomainName")

    @property
    def regional_domain_name(self):
        return GetAtt(name_to_id(self.name), "RegionalDomainName")

    def cfn_policy_document(self, stack: Stack) -> PolicyDocument:
        return PolicyDocument(
            [
                PolicyStatement(
                    action=[
                        "s3:CreateBucket",
                        "s3:DeleteBucket",
                        "s3:DeleteBucketPolicy",
                        "s3:GetBucketPolicy",
                        "s3:PutBucketPolicy",
                        "s3:PutEncryptionConfiguration",
                        "s3:GetEncryptionConfiguration",
                        "s3:PutBucketVersioning",
                        "s3:GetBucketVersioning",
                        "s3:PutBucketPublicAccessBlock",
                        "s3:GetBucketPublicAccessBlock",
                    ],
                    effect="Allow",
                    resource=self.arn,
                )
            ]
        )



from __future__ import annotations
from typing import TYPE_CHECKING
import argparse
import boto3
import botocore.session
import json
import logging
import os
import re
import requests
import requests.auth
import urllib.parse

from botocore.auth import SigV4Auth
from botocore.awsrequest import AWSRequest
from botocore.stub import Stubber
from uuid import uuid4


from e3.error import E3Error
from e3.env import Env
from e3.os.process import Run

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from typing import Any, TypedDict, Callable
    import botocore.client
    import botocore.stub

    class AWSCredentials(TypedDict, total=False):
        """Annotate a dict containing AWS credentials.

        :cvar Version: credentials format version
        :cvar AccessKeyId: access key ID that identifies the temporary security
            credentials
        :cvar SecretAccessKey: secret access key that can be used to sign requests
        :cvar SessionToken: token that users must pass to the service API to use
            the temporary credentials
        :cvar Expiration: date on which the current credentials expire
        """

        Version: int
        AccessKeyId: str
        SecretAccessKey: str
        SessionToken: str
        Expiration: str


class AWSSessionRunError(E3Error):
    def __init__(self, message: str, origin: str, process: Run | None = None) -> None:
        """Initialize an AWSSessionRunError.

        :param message: the exception message
        :param origin: the name of the function, class, or module having raised
        :param process: process that failed
        """
        super().__init__(message, origin)
        self.origin = origin
        self.message = message
        self.process = process


class Session:
    """Handle AWS session and clients."""

    def __init__(
        self,
        regions: list[str] | None = None,
        stub: bool = False,
        profile: str | None = None,
        credentials: AWSCredentials | None = None,
    ) -> None:
        """Initialize an AWS session.

        Once initialized AWS environment can be accessed from Env().aws_env

        :param regions: list of regions to work on. The first region is
            considered as the default region. This parameter should be provided
            if AWS environment variables are not used to specified the region
        :param stub: if True clients are necessarily stubbed
        :param profile: profile name
        :param credentials: AWS credentials dictionary containing the
            following keys: AccessKeyId, SecretAccessKey, SessionToken
            as returned by ``assume_role``
        """
        if profile is not None or credentials is None:
            self.session = botocore.session.Session(profile=profile)
        else:
            self.session = botocore.session.Session()
            self.session.set_credentials(
                access_key=credentials["AccessKeyId"],
                secret_key=credentials["SecretAccessKey"],
                token=credentials["SessionToken"],
            )

        self.profile = profile
        if regions is None:
            # the value return below is ('region', 'AWS_DEFAULT_REGION', None, None)
            # See botocore/configprovider.py
            region_variable = self.session.SESSION_VARIABLES["region"][1]
            region = os.environ.get(region_variable, "")
            if not region:
                raise ValueError(
                    "region should be specified either using regions "
                    "parameter or using AWS environment variables"
                )
            self.regions = [region]
        else:
            self.regions = regions

        self.default_region = self.regions[0]

        self.force_stub = stub
        self.clients: dict[str, botocore.client.Client] = {}
        self.stubbers: dict[str, botocore.stub.Stubber] = {}

        self._account_alias = None

        self._identity = None

    def assume_role(
        self,
        role_arn: str,
        role_session_name: str,
        session_duration: int | None = None,
    ) -> Session:
        """Return a session with ``role_arn`` credentials.

        :param role_arn: ARN of the role to assume
        :param role_session_name: a name to associate with the created
            session
        :param session_duration: session duration in seconds or None for
            default
        :return: a Session instance
        :rtype: Session
        """
        credentials = self.assume_role_get_credentials(
            role_arn, role_session_name, session_duration=session_duration
        )
        return Session(regions=self.regions, credentials=credentials)

    def assume_role_get_credentials(
        self,
        role_arn: str,
        role_session_name: str,
        session_duration: int | None = None,
        as_env_var: bool = False,
    ) -> AWSCredentials:
        """Return credentials for ``role_arn``.

        :param role_arn: ARN of the role to assume
        :param role_session_name: a name to associate with the created
            session
        :param session_duration: session duration in seconds or None for
            default
        :param as_env_var: if set to True the returned credentials dictionnary
            keys are translated to be compatible to update os.environ.
        """
        client = self.client("sts", region=self.regions[0])
        arguments: dict[str, Any] = {
            "RoleArn": role_arn,
            "RoleSessionName": role_session_name,
        }
        if session_duration is not None:
            arguments["DurationSeconds"] = session_duration

        # If AWS_MFA_DEVICE is in the environment ask user for OTP
        if "AWS_MFA_DEVICE" in os.environ:  # all: no cover
            arguments["SerialNumber"] = os.environ["AWS_MFA_DEVICE"]
            otp = input("Enter MFA code: ")
            arguments["TokenCode"] = otp

        response = client.assume_role(**arguments)

        credentials = response["Credentials"]
        if as_env_var:
            key_to_envvar = {
                "AccessKeyId": "AWS_ACCESS_KEY_ID",
                "SecretAccessKey": "AWS_SECRET_ACCESS_KEY",
                "SessionToken": "AWS_SESSION_TOKEN",
            }
            credentials = {
                key_to_envvar[k]: v
                for k, v in credentials.items()
                if k in key_to_envvar
            }

        return credentials

    @property
    def account_alias(self):
        """Return current account alias."""
        if self._account_alias is None:
            client = self.client("iam", region="us-east-1")
            aliases = client.list_account_aliases()["AccountAliases"]
            if aliases:
                # Even if the API return a list there currently only one
                # alias possible
                self._account_alias = aliases[0]
            else:
                self._account_alias = ""
        return self._account_alias

    @property
    def identity(self):
        """Return identity information."""
        if self._identity is None:
            sts_client = self.client("sts")
            self._identity = {"UserId": "", "Account": "", "Arn": ""}
            self._identity.update(sts_client.get_caller_identity())

        return self._identity

    def stub(self, name: str, region: str | None = None) -> botocore.stub.Stubber:
        """Return stub for a given client.

        Note that if the client does not exist yet it will be created.

        :param name: client name
        :param region: region associated with the client. If None the default
            region is taken.
        :return: the stub instance
        """
        if not self.force_stub:
            return None
        if region is None:
            region = self.default_region

        if name not in self.stubbers or region not in self.stubbers[name]:
            # Create client
            self.client(name, region)

        return self.stubbers[name][region]

    def to_boto3(self) -> boto3.Session:
        """Return boto3 session initialized from current botocore session."""
        credentials = self.session.get_credentials()
        frozen_credentials = credentials.get_frozen_credentials()
        return boto3.Session(
            aws_access_key_id=frozen_credentials.access_key,
            aws_secret_access_key=frozen_credentials.secret_key,
            aws_session_token=frozen_credentials.token,
            region_name=self.default_region,
            profile_name=self.profile,
        )

    def client(self, name: str, region: str | None = None) -> botocore.client.Client:
        """Get a client.

        :param name: client name
        :param region: region associated with the client. If None the default
            region is taken.
        :return: a client instance
        """
        if region is None:
            region = self.default_region

        assert region is not None, "no region or default_region set"

        if name not in self.clients:
            self.clients[name] = {}
            self.stubbers[name] = {}

        if region not in self.clients[name]:
            self.clients[name][region] = self.session.create_client(
                name, region_name=region
            )
            if self.force_stub:
                self.stubbers[name][region] = Stubber(self.clients[name][region])
                self.stubbers[name][region].activate()

        return self.clients[name][region]

    def run(
        self, cmd: list[str], role_arn: str, session_duration: int, **kwargs: Any
    ) -> Run:
        """Execute a command with credentials to assume role role_arn.

        :param cmd: command to execute
        :role_arn: Arn of the role to be used by the command
        :session_duration: session duration in seconds or None for default
        :param kwargs: additional parameters to provide to e3.os.process.Run
        :return: Result of the call to Run for the command
        """
        credentials = self.assume_role_get_credentials(
            role_arn, "aws_run_session", session_duration, as_env_var=True
        )

        if "env" not in kwargs:
            if "ignore_environ" not in kwargs:
                kwargs["ignore_environ"] = False
            kwargs["env"] = credentials
        else:
            kwargs["env"] = dict(kwargs["env"]).update(credentials)

        aws_p = Run(cmd, **kwargs)

        if aws_p.status:
            raise AWSSessionRunError(
                f"{cmd} failed (exit status: {aws_p.status})",
                origin="aws_session_cli_cmd",
                process=aws_p,
            )
        return aws_p


class AWSEnv(Session):
    """Handle AWS session and clients."""

    def __init__(
        self,
        regions: list[str] | None = None,
        stub: bool = False,
        profile: str | None = None,
    ):
        """Initialize an AWS session.

        Once intialized AWS environment can be accessed from Env().aws_env

        :param regions: list of regions to work on. The first region is
            considered as the default region.
        :param stub: if True clients are necessarily stubbed
        :param profile: profile name
        """
        super().__init__(regions=regions, stub=stub, profile=profile)
        env = Env()
        env.aws_env = self


class default_region:
    """Context manager used to set a default region."""

    def __init__(self, region: str):
        """Initialize context manager.

        :param region: default region
        """
        aws_env = Env().aws_env

        self.previous_region = aws_env.default_region
        self.default_region = region

    def __enter__(self):
        Env().aws_env.default_region = self.default_region

    def __exit__(self, _type, _value, _tb):
        del _type, _value, _tb
        Env().aws_env.default_region = self.previous_region


def session() -> Callable:
    """Decorate a function to handle automatically AWS session retrieval.

    The function in input should take a mandatory argument called session.

    :param name: session name
    :type name: str
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            if "session" in kwargs:
                session = kwargs.get("session")
                del kwargs["session"]
            else:
                session = Env().aws_env
            return func(*args, session=session, **kwargs)

        return wrapper

    return decorator


def assume_profile_main():
    """Generate shell commands to set credentials for a profile."""
    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument(
        "--json", action="store_true", help="Output credentials as JSON"
    )
    argument_parser.add_argument("profile")
    args = argument_parser.parse_args()

    # Frozen credentials returns a tuple so we need to convert it to
    # the same dict as in assume_role_main
    var_to_key = {
        "access_key": "AccessKeyId",
        "secret_key": "SecretAccessKey",
        "token": "SessionToken",
    }

    key_to_envvar = {
        "AccessKeyId": "AWS_ACCESS_KEY_ID",
        "SecretAccessKey": "AWS_SECRET_ACCESS_KEY",
        "SessionToken": "AWS_SESSION_TOKEN",
    }

    frozen_credentials = (
        boto3.Session(
            region_name="eu-west-1",
            profile_name=args.profile,
        )
        .get_credentials()
        .get_frozen_credentials()
    )

    credentials = {v: getattr(frozen_credentials, k) for k, v in var_to_key.items()}

    if args.json:
        print(json.dumps(credentials))
    else:
        credentials = {
            key_to_envvar[k]: v for k, v in credentials.items() if k in key_to_envvar
        }
        for k, v in credentials.items():
            print(f"export {k}={v}")


def assume_role_main():
    """Generate shell commands to set credentials for a role."""
    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument(
        "--json", action="store_true", help="Output credentials as JSON"
    )
    argument_parser.add_argument(
        "--role-session-name",
        help="Role session name, by default a random value is generated",
    )
    argument_parser.add_argument("--session-duration")
    argument_parser.add_argument("role_arn")
    args = argument_parser.parse_args()

    key_to_envvar = {
        "AccessKeyId": "AWS_ACCESS_KEY_ID",
        "SecretAccessKey": "AWS_SECRET_ACCESS_KEY",
        "SessionToken": "AWS_SESSION_TOKEN",
    }

    s = Session(regions=["eu-west-1"])
    session_duration = args.session_duration
    if session_duration is not None:
        session_duration = int(session_duration)

    if args.role_session_name:
        role_session_name = args.role_session_name
    else:
        role_session_name = str(uuid4()).replace("-", "")

    credentials = s.assume_role_get_credentials(
        args.role_arn, role_session_name, session_duration=session_duration
    )
    credentials["Expiration"] = credentials["Expiration"].timestamp()
    if args.json:
        print(json.dumps(credentials))
    else:
        credentials = {
            key_to_envvar[k]: v for k, v in credentials.items() if k in key_to_envvar
        }
        for k, v in credentials.items():
            print(f"export {k}={v}")


def iterate(fun: Callable, key: str, **kwargs: Any) -> Any:
    """Create an iterator other paginate botocore function.

    :param fun: the function to call
    :param key: the key in the returned data containing the elements
    :param kwargs: parameters passed to the function
    """
    result = fun(**kwargs)
    for data in result.get(key, []):
        yield data

    while result.get("NextToken"):
        result = fun(NextToken=result["NextToken"], **kwargs)
        for data in result.get(key, []):
            yield data


def name_to_id(name: str) -> str:
    """Convert a resource name to a resource id.

    The conversion is done by removing characters that are not alphanumeric as
    resource id must contain only alphanumeric character and first character and
    characters following a dash to uppercase for a better readability.
    """

    def replacement(match):
        """Return uppercased second character of the match."""
        return match.group(1)[1].upper()

    resource_id = re.sub(r"[^a-zA-Z0-9]", "", re.sub(r"(-[a-z])", replacement, name))
    resource_id = resource_id[0].upper() + resource_id[1:]
    return resource_id


class IAMAuth(requests.auth.AuthBase):
    """Authorizer for the requests framework that use AWS Signature V4 protocol."""

    def __init__(
        self, session: Session, role: str | None = None, region: str | None = None
    ):
        """Initialize authorizer.

        :param session: an aws session
        :param role: if not None role to assume
        :param region: if None use default region (from session) otherwise use it as
            region
        """
        self.session = session
        self.role = role
        if region is None:
            self.region = self.session.default_region
        else:
            self.region = region

    def __call__(self, request: requests.PreparedRequest) -> requests.PreparedRequest:
        """See requests framework."""
        # The code does not implement the signature v4 protocol. It reuses parts of
        # botocore to do so. Botocore does not provide a generic interface to sign
        # a given request. So the following code create a temporary AWSRequest object
        # with the same parameters as the user request, sign it and then extract
        # the signature (i.e: the headers) that is re-injected into the user request.
        session = self.session
        if self.role is not None:
            session = self.session.assume_role(self.role, "iamauthsession")

        credentials = session.session.get_credentials().get_frozen_credentials()

        # Split back the url in order to be able to call AWSRequest
        aws_headers = dict(request.headers)
        key_to_delete = []
        for key in aws_headers:
            if key.lower() in ("accept", "accept-encoding", "connection"):
                key_to_delete.append(key)

        for key in key_to_delete:
            del aws_headers[key]

        parsed_url = urllib.parse.urlparse(request.url)
        aws_params = {
            str(k): ",".join(v)
            for k, v in urllib.parse.parse_qs(str(parsed_url.query)).items()
        }
        aws_url = urllib.parse.urlunparse(
            (
                str(parsed_url.scheme),
                str(parsed_url.netloc),
                str(parsed_url.path),
                str(parsed_url.params),
                "",
                "",
            )
        )

        # Compute the headers for the request
        aws_request = AWSRequest(
            method=request.method, url=aws_url, params=aws_params, headers=aws_headers
        )
        SigV4Auth(credentials, "execute-api", self.region).add_auth(aws_request)

        # Update request headers
        request.headers.update(aws_request.headers)
        return request



from __future__ import annotations
from abc import ABC, abstractmethod
from itertools import chain
from troposphere import AWSObject, Template

from e3.aws import cfn, name_to_id, Session
from e3.aws.cfn.main import CFNMain
from e3.aws.troposphere.iam.policy_document import PolicyDocument
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # all: no cover
    from typing import Union
    from troposphere import And, Condition, Equals, If, Not, Or

    ConditionFunction = Union[And, Condition, Equals, If, Not, Or]


class Construct(ABC):
    """Represent one or multiple troposphere AWSObject.

    AWSObjects are accessible with resources attribute.
    """

    @abstractmethod
    def resources(self, stack: Stack) -> list[AWSObject | Construct]:
        """Return a list of troposphere AWSObject.

        Objects returned can be added to a troposphere template with
        add_resource Template method.

        :param stack: the stack that contains the construct
        """
        pass

    def cfn_policy_document(self, stack: Stack) -> PolicyDocument:
        """Return the IAM policy needed by CloudFormation to manage the stack.

        :param stack: the stack that contains the construct
        """
        return PolicyDocument([])

    def create_data_dir(self, root_dir: str) -> None:  # noqa: B027
        """Put data in root_dir before export to S3 bucket referenced by the stack.

        :param root_dir: local directory in which data should be stored. Data will
            be then uploaded to an S3 bucket accessible from the template. The
            target location is the one received by resources method. Note that
            the same root_dir is shared by all resources in your stack.
        """
        pass


class Stack(cfn.Stack):
    """Cloudformation stack using troposphere resources."""

    def __init__(
        self,
        stack_name: str,
        description: str | None = None,
        cfn_role_arn: str | None = None,
        deploy_session: Session | None = None,
        dry_run: bool | None = False,
        s3_bucket: str | None = None,
        s3_key: str | None = None,
        version: str | None = None,
    ) -> None:
        """Initialize Stack attributes.

        :param stack_name: stack name
        :param cfn_role_arn: role asssumed by cloud formation to create the stack
        :param deploy_session: AWS session to deploy non CloudFormation AWS
            resources (aka Assets)
        :param dry_run: True if the stack is not to be deployed.
        :param description: a description of the stack
        :param s3_bucket: s3 bucket used to store data needed by the stack
        :param s3_key: s3 prefix in s3_bucket in which data is stored
        :param version: template format version
        """
        super().__init__(
            stack_name,
            cfn_role_arn=cfn_role_arn,
            description=description,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
        )
        self.constructs: list[Construct | AWSObject] = []

        self.deploy_session = deploy_session
        self.dry_run = dry_run
        self.version = version
        self.template = Template()

    def construct_to_objects(self, construct: Construct | AWSObject) -> list[AWSObject]:
        """Return list of AWS objects resources from a construct.

        :param construct: construct to list resources from
        """
        if isinstance(construct, AWSObject):
            return [construct]
        else:
            return list(
                chain.from_iterable(
                    [
                        self.construct_to_objects(el)
                        for el in construct.resources(stack=self)
                    ]
                )
            )

    def add(self, element: AWSObject | Construct | Stack) -> Stack:
        """Add a Construct or AWSObject to the stack.

        :param element: if a resource an AWSObject or Construct add the resource
             to the stack. If a stack merge its resources into the current stack.
        """
        if isinstance(element, Stack):
            constructs = element.constructs

        else:
            constructs = [element]

        # Add the new constructs (non expanded)
        self.constructs += constructs

        # Update the template
        resources = []
        for construct in constructs:
            resources.extend(self.construct_to_objects(construct))
        self.template.add_resource(resources)

        return self

    def add_condition(self, condition_name: str, condition: ConditionFunction) -> None:
        """Add condition to stack template.

        :param condition_name: name of the condition to add
        :param condition: condition to add
        """
        self.template.add_condition(condition_name, condition)

    def cfn_policy_document(self) -> PolicyDocument:
        """Return stack necessary policy document for CloudFormation."""
        result = PolicyDocument([])
        for construct in self.constructs:
            if isinstance(construct, Construct):
                result += construct.cfn_policy_document(stack=self)

        return result

    def __getitem__(self, resource_name: str) -> AWSObject:
        """Return AWSObject associated with resource_name.

        :param resource_name: name of the resource to retrieve
        """
        return self.template.resources[name_to_id(resource_name)]

    def export(self) -> dict:
        """Export stack as dict.

        :return: a dict that can be serialized as YAML to produce a template
        """
        result = self.template.to_dict()
        result["AWSTemplateFormatVersion"] = (
            "2010-09-09" if self.version is None else self.version
        )
        if self.description is not None:
            result["Description"] = self.description
        return result

    def create_data_dir(self, root_dir: str) -> None:
        """Populate root_dir with data needed by all constructs in the stack.

        :param root_dir: the local directory in which to store the data
        """
        for construct in self.constructs:
            if isinstance(construct, Construct):
                construct.create_data_dir(root_dir)


class CFNProjectMain(CFNMain):
    """CFNMain with default value and self initializing its stack.

    This facilitates the deployment of mono stack projects with deployment and
    CloudFormation roles named respectively cfn-user/CFNAllowDeployOf<stack_name>
    and cfn-service/CFNServiceRoleFor<stack_name>.
    """

    def __init__(
        self,
        name: str,
        account_id: str,
        stack_description: str,
        s3_bucket: str,
        regions: list[str],
    ) -> None:
        """
        Initialize a CFNProjectMain instance.

        :param name: name of the project
        :param account_id: id of the account where to deploy the project
        :param stack_description: description of the stack to deploy
        :param s3_bucket: see CFNMain
        :param regions: see CFNMain
        """
        super().__init__(
            regions=regions,
            s3_bucket=s3_bucket,
            s3_key=name,
            assume_role=(
                f"arn:aws:iam::{account_id}:role/cfn-user/CFNAllowDeployOf{name}",
                f"Deploy{name}Session",
            ),
        )
        self.stack = Stack(
            name,
            cfn_role_arn=f"arn:aws:iam::{account_id}:role/cfn-service/CFNServiceRoleFor{name}",
            description=stack_description,
            s3_bucket=s3_bucket,
            s3_key=self.s3_data_key,
        )

    def add(self, element: AWSObject | Construct | Stack) -> Stack:
        """Add resource to project's stack.

        :param element: resource to add to the stack.
        """
        return self.stack.add(element)



"""Ease deployment of a lambda used in API Gateway using Flask framework.

For example if you have app.py in /myapp directory::

    from flask import Flask
    app = Flask(__name__)

    @app.route('/')
    def hello_world():
        return 'Hello, World!'

You can create the associated lambda by adding the following element in your
stack::

    lambda_function = PyFlaskFunction(name="myapp",
                                      description="a comment",
                                      role=lambda_role,
                                      code_dir="/myapp",
                                      app="app.app",
                                      runtime="python3.9")
    stack.add(lambda_function)

The resulting lambda can then be added into an HttpAPI::

    stack.add(HttpApi(name="myapi",
                      description="my api description",
                      lambda_arn=lambda_function.arn,
                      route_list=[Get("/")]))

"""
from __future__ import annotations
import os

from troposphere import GetAtt
from e3.aws.troposphere.iam.role import Role
from e3.fs import cp, sync_tree
from . import PyFunction


STARTUP_CODE = """
from %(app_module)s import %(app_name)s
from flask_apigateway_wrapper import FlaskLambdaHandler

lambda_handler = FlaskLambdaHandler(%(app_name)s)
lambda_handler_fun = lambda_handler.lambda_handler
"""


class PyFlaskFunction(PyFunction):
    """Provide PyFlaskFunction Construct."""

    def __init__(
        self,
        name: str,
        description: str,
        role: str | GetAtt | Role,
        code_dir: str,
        app: str,
        runtime: str,
        requirement_file: str | None = None,
        code_version: int | None = None,
        timeout: int = 3,
        memory_size: int | None = None,
        logs_retention_in_days: int | None = 731,
        reserved_concurrent_executions: int | None = None,
        environment: dict[str, str] | None = None,
    ):
        """Initialize a Flask AWS lambda function using a Python runtime.

        :param name: function name
        :param description: a description of the function
        :param role: role to be asssumed during lambda execution
        :param code_dir: directory containing the python code
        :param app: Flask app symbol name
        :param runtime: lambda runtime. It must be a Python runtime.
        :param requirement_file: requirement file for the application code.
            Required packages are automatically fetched (works only from linux)
            and packaged along with the lambda code
        :param code_version: code version
        :param timeout: maximum execution time (default: 3s)
        :param memory_size: the amount of memory available to the function at
            runtime. The value can be any multiple of 1 MB.
        :param logs_retention_in_days: The number of days to retain the log events
            in the lambda log group
        :param reserved_concurrent_executions: The number of concurrent executions
            that are reserved for this function
        :param environment: Environment variables that are accessible from function
            code during execution
        """
        self.app_module, self.app_name = app.rsplit(".", 1)

        super().__init__(
            name=name,
            description=description,
            role=role,
            code_dir=code_dir,
            handler="lambda_handler_module.lambda_handler_fun",
            runtime=runtime,
            requirement_file=requirement_file,
            code_version=code_version,
            timeout=timeout,
            memory_size=memory_size,
            logs_retention_in_days=logs_retention_in_days,
            reserved_concurrent_executions=reserved_concurrent_executions,
            environment=environment,
        )

    def populate_package_dir(self, package_dir: str) -> None:
        super().populate_package_dir(package_dir=package_dir)

        handler_file = os.path.join(package_dir, "lambda_handler_module.py")
        with open(handler_file, "w") as fd:
            fd.write(
                STARTUP_CODE
                % {"app_module": self.app_module, "app_name": self.app_name}
            )
        wrapper_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "flask_apigateway_wrapper.py",
        )
        cp(wrapper_file, package_dir)
        sync_tree(package_dir, "/tmp/nico/package")


class Py38FlaskFunction(PyFlaskFunction):
    """Provide Py38FlaskFunction Construct."""

    def __init__(
        self,
        name: str,
        description: str,
        role: str | GetAtt | Role,
        code_dir: str,
        app: str,
        requirement_file: str | None = None,
        code_version: int | None = None,
        timeout: int = 3,
        memory_size: int | None = None,
        logs_retention_in_days: int | None = None,
    ):
        """Initialize a Flask AWS lambda function using Python 3.8 runtime.

        For params see PyFlaskFunction.
        """
        super().__init__(
            name=name,
            description=description,
            role=role,
            code_dir=code_dir,
            app=app,
            runtime="python3.8",
            requirement_file=requirement_file,
            code_version=code_version,
            timeout=timeout,
            memory_size=memory_size,
            logs_retention_in_days=logs_retention_in_days,
        )



"""Provide s3 high level constructs."""
from __future__ import annotations

from typing import TYPE_CHECKING

from e3.aws.troposphere import Construct
from e3.aws.troposphere.iam.managed_policy import ManagedPolicy
from e3.aws.troposphere.iam.policy_statement import Allow, Trust
from e3.aws.troposphere.iam.role import Role
from e3.aws.troposphere.s3.bucket import Bucket

if TYPE_CHECKING:
    from typing import Any
    from troposphere import AWSObject, Stack


class BucketWithRoles(Construct):
    """Provide resources for a s3 bucket with its access roles."""

    def __init__(
        self,
        name: str,
        iam_names_prefix: str,
        iam_path: str,
        trusted_accounts: list[str],
        iam_read_root_name: str = "Read",
        iam_write_root_name: str = "Write",
        **bucket_kwargs: Any,
    ) -> None:
        """Initialize BucketWithRoles instance.

        :param name: name of the bucket
        :param iam_names_prefix: prefix for policies and roles names
        :param iam_path: path for iam resources
        :param trusted_accounts: accounts to be trusted by access roles
        :param iam_read_root_name: root name for read access roles and policy
        :param iam_write_root_name: root name for write access roles and policy
        :param bucket_kwargs: keyword arguments to pass to the bucket constructor
        """
        self.name = name
        self.iam_names_prefix = iam_names_prefix
        self.trusted_accounts = trusted_accounts

        self.bucket = Bucket(name=self.name, **bucket_kwargs)
        self.read_policy = ManagedPolicy(
            name=f"{self.iam_names_prefix}{iam_read_root_name}Policy",
            description=f"Grants read access permissions to {self.name} bucket",
            statements=[
                Allow(action=["s3:GetObject"], resource=self.bucket.all_objects_arn),
                Allow(action=["s3:ListBucket"], resource=self.bucket.arn),
            ],
            path=iam_path,
        )
        self.read_role = Role(
            name=f"{self.iam_names_prefix}{iam_read_root_name}Role",
            description=f"Role with read access to {self.name} bucket.",
            trust=Trust(accounts=self.trusted_accounts),
            managed_policy_arns=[self.read_policy.arn],
            path=iam_path,
        )
        self.push_policy = ManagedPolicy(
            name=f"{self.iam_names_prefix}{iam_write_root_name}Policy",
            description=f"Grants write access permissions to {self.name} bucket",
            statements=[
                Allow(
                    action=["s3:PutObject", "s3:DeleteObject"],
                    resource=self.bucket.all_objects_arn,
                )
            ],
            path=iam_path,
        )
        self.push_role = Role(
            name=f"{self.iam_names_prefix}{iam_write_root_name}Role",
            description=f"Role with read and write access to {self.name} bucket.",
            trust=Trust(accounts=self.trusted_accounts),
            managed_policy_arns=[self.push_policy.arn, self.read_policy.arn],
            path=iam_path,
        )

    @property
    def ref(self):
        """Return bucket ref."""
        return self.bucket.ref

    @property
    def arn(self):
        """Return bucket arn."""
        return self.bucket.arn

    @property
    def all_objects_arn(self):
        return self.bucket.all_objects_arn

    def resources(self, stack: Stack) -> list[AWSObject | Construct]:
        """Return resources associated with the construct."""
        return [
            self.bucket,
            self.read_policy,
            self.read_role,
            self.push_policy,
            self.push_role,
        ]



from __future__ import annotations
import os

from troposphere import Ref
from e3.aws.troposphere import Stack
from e3.aws.troposphere.dynamodb import Table


SOURCE_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "source_dir")


EXPECTED_TABLE_DEFAULT_TEMPLATE = {
    "Mytable": {
        "Properties": {
            "AttributeDefinitions": [{"AttributeName": "id", "AttributeType": "N"}],
            "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
            "PointInTimeRecoverySpecification": {"PointInTimeRecoveryEnabled": True},
            "ProvisionedThroughput": {
                "ReadCapacityUnits": 10,
                "WriteCapacityUnits": 10,
            },
            "TableName": "mytable",
        },
        "Type": "AWS::DynamoDB::Table",
    },
}


EXPECTED_TABLE_TEMPLATE = {
    "Mytable": {
        "Properties": {
            **EXPECTED_TABLE_DEFAULT_TEMPLATE["Mytable"]["Properties"],
            **{
                "Tags": [{"Key": "tagkey", "Value": "tagvalue"}],
                "TimeToLiveSpecification": {
                    "AttributeName": "ExpirationTime",
                    "Enabled": True,
                },
                "BillingMode": "PROVISIONED",
                "ProvisionedThroughput": {
                    "ReadCapacityUnits": 20,
                    "WriteCapacityUnits": {"Ref": "WriteCapacityUnits"},
                },
                "StreamSpecification": {"StreamViewType": "NEW_IMAGE"},
            },
        },
        "Type": "AWS::DynamoDB::Table",
    },
}


def test_table_default(stack: Stack) -> None:
    """Test default Table creation."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"
    stack.add(
        Table(
            name="mytable", attribute_definitions={"id": "N"}, key_schema={"id": "HASH"}
        )
    )
    assert stack.export()["Resources"] == EXPECTED_TABLE_DEFAULT_TEMPLATE


def test_table(stack: Stack) -> None:
    """Test Table creation."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"
    stack.add(
        Table(
            name="mytable",
            attribute_definitions={"id": "N"},
            key_schema={"id": "HASH"},
            tags={"tagkey": "tagvalue"},
            point_in_time_recovery_enabled=True,
            billing_mode="PROVISIONED",
            read_capacity_units=20,
            write_capacity_units=Ref("WriteCapacityUnits"),
            time_to_live_attribute_name="ExpirationTime",
            time_to_live_enabled=True,
            stream_enabled=True,
            stream_view_type="NEW_IMAGE",
        )
    )
    assert stack.export()["Resources"] == EXPECTED_TABLE_TEMPLATE



"""Provide Event Rules constructs."""
from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING


from troposphere import AWSObject, events, GetAtt, Ref

from e3.aws import name_to_id
from e3.aws.troposphere import Construct
from e3.aws.troposphere.ec2 import VPC
from e3.aws.troposphere.ecs.cluster import FargateCluster
from e3.aws.troposphere.ecs.vpc import EcsVPC

if TYPE_CHECKING:
    from e3.aws.troposphere import Stack


@dataclass(frozen=True)
class FargateScheduledTaskRule(Construct):
    """Define an EventBridge rule that schedule fargate tasks.

    :param description: rule description
    :param ecs_cluster: ECS Cluster that should run the scheduled task
    :param name: name of the rule
    :param schedule_expression: expression that defines when the rule is
        triggered. For example to trigger the rule every 15 minutes the
        following expression can be given "cron(0/15 * * * ? *)".
    :param task_names: List of tasks to schedule
    :param vpc: VPC used to run the scheduled task
    :param state: state of the rule (DISABLED | ENABLED)
    """

    description: str
    ecs_cluster: FargateCluster
    name: str
    schedule_expression: str
    task_names: list[str]
    vpc: EcsVPC | VPC
    state: str = "DISABLED"

    def ecs_parameters(self, task_name: str) -> events.EcsParameters:
        """Return ECS parameters describing the fargate task to run.

        :param task_name: name of the task
        """
        if isinstance(self.vpc, EcsVPC):
            subnet = self.vpc.subnet
        else:
            subnet = self.vpc.main_subnet

        return events.EcsParameters(
            LaunchType="FARGATE",
            NetworkConfiguration=events.NetworkConfiguration(
                AwsVpcConfiguration=events.AwsVpcConfiguration(
                    AssignPublicIp="DISABLED",
                    SecurityGroups=[Ref(self.vpc.security_group)],
                    Subnets=[Ref(subnet)],
                )
            ),
            TaskDefinitionArn=Ref(name_to_id(f"{task_name}")),
            PlatformVersion="1.4.0",
        )

    @property
    def targets(self) -> list[events.Target]:
        """Return rule's targets."""
        return [
            events.Target(
                Arn=GetAtt(name_to_id(self.ecs_cluster.name), "Arn"),
                RoleArn=GetAtt(self.ecs_cluster.ecs_events_role.name, "Arn"),
                EcsParameters=self.ecs_parameters(task_name),
                Id=name_to_id(f"{task_name}-target"),
            )
            for task_name in self.task_names
        ]

    @property
    def rule(self) -> events.Rule:
        """Return the rule scheduling the fargate task."""
        return events.Rule(
            name_to_id(self.name),
            Description=self.description,
            Name=self.name,
            ScheduleExpression=self.schedule_expression,
            State=self.state,
            Targets=self.targets,
        )

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return FargateScheduledRule resources."""
        return [self.rule]



from __future__ import annotations

from e3.aws.troposphere import Stack
from e3.aws.troposphere.sns import Topic


EXPECTED_TOPIC_DEFAULT_TEMPLATE = {
    "Mytopic": {
        "Properties": {"Subscription": [], "TopicName": "mytopic"},
        "Type": "AWS::SNS::Topic",
    }
}


EXPECTED_TOPIC_TEMPLATE = {
    "Mytopic": {
        "Properties": {
            "Subscription": [],
            "TopicName": "mytopic",
            "KmsMasterKeyId": "arn:aws:kms:eu-west-1:012345678901:key/"
            "00000000-0000-0000-0000-000000000000",
        },
        "Type": "AWS::SNS::Topic",
    }
}


def test_topic_default(stack: Stack) -> None:
    """Test Topic default creation."""
    stack.add(Topic(name="mytopic"))
    assert stack.export()["Resources"] == EXPECTED_TOPIC_DEFAULT_TEMPLATE


def test_topic(stack: Stack) -> None:
    """Test Topic creation."""
    stack.add(
        Topic(
            name="mytopic",
            kms_master_key_id="arn:aws:kms:eu-west-1:012345678901:key/"
            "00000000-0000-0000-0000-000000000000",
        )
    )
    assert stack.export()["Resources"] == EXPECTED_TOPIC_TEMPLATE



#!/usr/bin/env python
from __future__ import annotations
from argparse import ArgumentParser
import os

from e3.env import Env

from configparser import ConfigParser


def main(coverage_rc: str, omit_list_filename: str | None = None) -> None:
    os_name = Env().build.os.name
    test_dir = os.path.abspath(os.path.dirname(__file__))

    config = ConfigParser()
    base_conf, target_conf = (
        os.path.join(test_dir, "coverage", "%s.rc" % name) for name in ("base", os_name)
    )

    with open(coverage_rc, "w") as dest:
        config.read(base_conf)
        config.read(target_conf)

        # exclude lines is built with: base.rc config
        exclude_lines = config.get("report", "exclude_lines").splitlines()

        # add all <os>-only patterns
        exclude_lines += [
            "%s-only" % o
            for o in ("darwin", "linux", "solaris", "windows", "bsd", "aix")
            if o != os_name
        ]
        # exclude this specific os
        exclude_lines.append("%s: no cover" % os_name)

        # special case for unix
        if os_name != "windows":
            exclude_lines.append("unix: no cover")

        config.set("report", "exclude_lines", "\n".join(exclude_lines))

        # If the user gave a file with a list of "omit" entries,
        # read it, and append its contents to the run.omit option.

        if omit_list_filename is not None:
            if config.has_option("run", "omit"):
                omit = config.get("run", "omit").splitlines()
            else:
                omit = []
            with open(omit_list_filename) as f:
                for line in f:
                    omit.append(line)
            config.set("run", "omit", "\n".join(omit))

        config.write(dest)


if __name__ == "__main__":
    parser = ArgumentParser(description="Generate a coverage config file")
    parser.add_argument(
        "coverage_rc_filename", help="The name of the coverage configuration file."
    )
    parser.add_argument(
        "--omit-from-file",
        dest="omit_list_filename",
        help=(
            "The name of the file providing a list of files which should"
            " be excluded in the coverage report, with each line being"
            " a glob pattern matching the files that should be omitted."
            " This omit list is in addition to the list of files to"
            " be omitted by default."
        ),
    )

    args = parser.parse_args()
    main(args.coverage_rc_filename, omit_list_filename=args.omit_list_filename)



import re
from datetime import datetime

from dateutil.parser import parse as parse_date
from e3.aws import session
from e3.aws.ec2 import BlockDeviceMapping, EC2Element


class AMI(EC2Element):
    """Represent an AMI."""

    PROPERTIES = {
        "ImageId": "id",
        "OwnerId": "owner_id",
        "Public": "public",
        "RootDeviceName": "root_device",
    }

    @session()
    def __init__(self, ami_id, region=None, data=None, session=None):
        """Inialize an AMI description object.

        :param ami_id: the id of the AMI
        :type ami_id: str
        :param region: region in which the AMI is present. If None then
            use default region
        :type region: None | str
        :param data: a dict representing the metadata of the AMI. If None then
            download AMI description using EC2 api
        :type data: dict | None
        """
        if data is None:
            assert ami_id is not None
            data = session.client("ec2", region).describe_images(ImageIds=[ami_id])[
                "Images"
            ][0]
        super().__init__(data, region)

    @property
    def creation_date(self):
        """Creation date.

        :return: AMI creation date
        :rtype: datetime.datetime
        """
        return parse_date(self.data["CreationDate"]).replace(tzinfo=None)

    @property
    def age(self):
        """Return age of the AMI in days."""
        age = datetime.now() - self.creation_date
        return int(age.total_seconds() / (3600 * 24))

    @property
    def os_version(self):
        return self.tags.get("os_version", "unknown")

    @property
    def platform(self):
        return self.tags.get("platform", "unknown")

    @property
    def kind(self):
        return self.tags.get("kind", "unknown")

    @property
    def is_windows(self):
        return "windows" in self.data.get("Platform", "unknown")

    @property
    def timestamp(self):
        return int(self.tags.get("timestamp", "0"))

    @property
    def block_device_mappings(self):
        return [
            BlockDeviceMapping(bdm, region=self.region)
            for bdm in self.data.get("BlockDeviceMappings", [])
        ]

    @property
    def snapshot_ids(self):
        result = []
        for device in self.block_device_mappings:
            if device.is_ebs:
                result.append(device.snapshot_id)
        return result

    def __str__(self):
        return "%-12s %-24s: %s" % (
            self.region,
            self.data["ImageId"],
            self.data.get("Description", ""),
        )

    @classmethod
    @session()
    def ls(cls, filters=None, session=None):
        """List user AMIs.

        :param filters: same as Filters parameters of describe_images
            (see botocore)
        :type filters: dict
        :return a list of images
        :rtype: list[AMI]
        """
        if filters is None:
            filters = []
        result = []
        for r in session.regions:
            c = session.client("ec2", r)
            region_result = c.describe_images(Owners=["self"], Filters=filters)
            for ami in region_result["Images"]:
                result.append(AMI(ami["ImageId"], r, data=ami, session=session))
        return result

    @classmethod
    @session()
    def find(
        cls,
        platform=None,
        os_version=None,
        kind=None,
        region=None,
        session=None,
        **kwargs
    ):
        """Find AMIs.

        Only AMIs with platform, timestamps, os_version are considered.

        If kind is not None only consider AMIs also having a kind tag.

        :param platform: platform to match. If None all platforms are matched
        :type platform: str | None
        :param os_version: os_version to match. If None all os_versions are
            matched
        :type os_version: str | None
        :param kind: kind to match. If None all regions are matched
        :type kind: str | None
        :param region: region to match. If None all regions are matched
        :type region: str | None
        :param kwargs: additional filters on tags. parameter name if the tag
            name and the associated value the regexp
        :type kwargs: dict
        :return: a list of AMI
        :rtype: list[AMI]
        """
        result = {}

        filters = [
            {"Name": "tag-key", "Values": ["platform"]},
            {"Name": "tag-key", "Values": ["timestamp"]},
            {"Name": "tag-key", "Values": ["os_version"]},
        ]
        if kind is not None:
            filters.append({"Name": "tag-key", "Values": ["kind"]})

        all_images = AMI.ls(filters=filters, session=session)

        tag_filters = dict(kwargs)
        if platform is not None:
            tag_filters["platform"] = platform
        if os_version is not None:
            tag_filters["os_version"] = os_version
        if kind is not None:
            tag_filters["kind"] = kind

        for ami in all_images:
            key_l = [ami.region, ami.platform, ami.os_version]
            if kind is not None:
                key_l.append(ami.kind)
            key = tuple(key_l)

            if region is not None and not re.match(region, ami.region):
                continue

            consider_ami = True
            for tag in tag_filters:
                if not re.match(tag_filters[tag], ami.tags.get(tag, "")):
                    consider_ami = False
                    continue

            if not consider_ami:
                continue

            if key not in result or result[key][0] < ami.timestamp:
                result[key] = (ami.timestamp, ami)

        return [el[1] for el in result.values()]

    @classmethod
    @session()
    def select(
        cls, platform, os_version, kind=None, region=None, session=None, **kwargs
    ):
        """Select one AMI based on platform, os_version and kind.

        :param platform: platform name
        :type platform: str
        :param os_version: OS version
        :type os_version: str
        :param kind: kind
        :type kind: str
        :param region: region name or None (default region)
        :type region: str | None
        :return: one AMI
        :rtype: AMI
        """
        if region is None:
            region = session.default_region

        kind_filter = kind + "$" if kind is not None else None
        result = AMI.find(
            platform=platform + "$",
            os_version=os_version + "$",
            kind=kind_filter,
            region=region,
            session=session,
            **kwargs
        )
        assert (
            len(result) == 1
        ), "cannot find AMI %s (%s) of kind (%s) in region %s %s" % (
            platform,
            os_version,
            kind,
            region,
            kwargs,
        )
        return result[0]



from __future__ import annotations
import abc
import logging
import os
import tempfile
import time
import json
from datetime import datetime

import botocore.exceptions

from e3.aws import AWSEnv, Session
from e3.aws.cfn import Stack
from e3.env import Env
from e3.fs import find, sync_tree
from e3.main import Main


class CFNMain(Main, metaclass=abc.ABCMeta):
    """Main to handle CloudFormation stack from command line."""

    def __init__(
        self,
        regions: list[str],
        default_profile: str | None = None,
        data_dir: str | None = None,
        s3_bucket: str | None = None,
        s3_key: str = "",
        assume_role: tuple[str, str] | None = None,
    ):
        """Initialize main.

        :param regions: list of regions on which we can operate
        :param default_profile: default AWS profile to use to create the stack
        :param data_dir: directory containing files used by cfn-init
        :param s3_bucket: if defined S3 will be used as a proxy for resources.
            Template body will be uploaded to S3 before calling operation on
            it. This change the body limit from 50Ko to 500Ko. Additionally if
            data_dir is defined, the directory will be uploaded to the
            specified S3 bucket.
        :param s3_key: if s3_bucket is defined, then all uploaded resources
            will be stored under a subkey of s3_key. If not defined the root
            of the bucket is used.
        :param assume_role: tuple containing the two values that are passed
            to Session.assume_role()
        """
        super(CFNMain, self).__init__(platform_args=False)
        self.argument_parser.add_argument(
            "--profile",
            help="choose AWS profile{}".format(
                "" if default_profile is None else f", default is {default_profile}"
            ),
            default=default_profile,
        )

        if len(regions) > 1:
            self.argument_parser.add_argument(
                "--region",
                help="choose region (default: %s)" % regions[0],
                default=regions[0],
            )
        else:
            self.argument_parser.set_defaults(region=regions[0])

        subs = self.argument_parser.add_subparsers(
            title="commands", description="available commands", dest="command"
        )
        subs.required = True

        create_args = subs.add_parser("push", help="push a stack")
        create_args.add_argument(
            "--no-wait",
            action="store_false",
            default=True,
            dest="wait_stack_creation",
            help="do not wait for stack creation completion",
        )
        create_args.set_defaults(command="push")

        update_args = subs.add_parser("update", help="update a stack")
        update_args.add_argument(
            "--no-apply",
            action="store_false",
            default=True,
            dest="apply_changeset",
            help="do not ask whether to apply the changeset",
        )
        update_args.add_argument(
            "--no-wait",
            action="store_false",
            default=True,
            dest="wait_stack_creation",
            help="do not wait for stack update completion",
        )
        update_args.add_argument(
            "-y",
            "--yes",
            action="store_true",
            dest="skip_prompts",
            help="automatic yes to prompts",
        )
        update_args.set_defaults(command="update")

        show_args = subs.add_parser("show", help="show the changeset content")
        show_args.set_defaults(command="show")

        protect_args = subs.add_parser(
            "protect", help="protect the stack against deletion"
        )
        protect_args.set_defaults(command="protect")

        delete_args = subs.add_parser("delete", help="delete stack")
        delete_args.add_argument(
            "--no-wait",
            action="store_false",
            default=True,
            dest="wait_stack_creation",
            help="do not wait for stack deletion completion",
        )
        delete_args.set_defaults(command="delete")

        show_cfn_policy_args = subs.add_parser(
            "show-cfn-policy", help="show required policy for CFN"
        )
        show_cfn_policy_args.set_defaults(command="show-cfn-policy")

        self.regions = regions

        self.data_dir = data_dir
        self.s3_bucket = s3_bucket
        self.s3_data_key = None
        self.s3_data_url = None
        self.s3_template_key = None
        self.s3_template_url = None
        self.assume_role = assume_role

        self.timestamp = datetime.utcnow().strftime("%Y-%m-%d/%H:%M:%S.%f")

        if s3_bucket is not None:
            s3_root_key = (
                "/".join([s3_key.rstrip("/"), self.timestamp]).strip("/") + "/"
            )
            self.s3_data_key = s3_root_key + "data/"
            self.s3_data_url = "https://%s.s3.amazonaws.com/%s" % (
                self.s3_bucket,
                self.s3_data_key,
            )
            self.s3_template_key = s3_root_key + "template"
            self.s3_template_url = "https://%s.s3.amazonaws.com/%s" % (
                self.s3_bucket,
                self.s3_template_key,
            )

    @property
    def dry_run(self) -> bool:
        """Return True if CloudFormation stack is not to be deployed."""
        assert self.args is not None
        return self.args.command not in ("push", "update")

    def create_data_dir(self, root_dir: str) -> None:
        """Sync into root_dir data uploaded to the s3 bucket used by the stack.

        By default the content of self.data_dir is copied into root_dir.
        self.s3_data_key and self.s3_bucket can be used to reference resources in
        the template. The method can be overriden to create dynamically the content
        to upload.

        :param root_dir: location in which data to upload should be placed.
        """
        if self.data_dir is not None:
            sync_tree(self.data_dir, root_dir)

    def _prompt_yes(self, msg: str) -> bool:
        """Prompt user for yes or no answer.

        :param msg: short question to ask user
        :return: if user answered yes
        """
        assert self.args is not None
        if self.args.skip_prompts:
            return True

        ask = input(f"{msg} (y/N): ")
        return ask[0] in "Yy"

    def execute_for_stack(self, stack: Stack, aws_env: Session | None = None) -> int:
        """Execute application for a given stack and return exit status.

        :param Stack: the stack on which the application executes
        :param aws_env: custom AWS session to use
        """
        assert self.args is not None
        try:
            if self.args.command != "show":
                self.start_session(
                    profile=self.args.profile, region=self.args.region, aws_env=aws_env
                )

            if self.args.command in ("push", "update"):
                # Synchronize resources to the S3 bucket
                s3 = self.aws_env.client("s3")
                with tempfile.TemporaryDirectory() as tempd:
                    # Push data associated with CFNMain and then all data
                    # related to the stack
                    self.create_data_dir(root_dir=tempd)
                    stack.create_data_dir(root_dir=tempd)

                    if self.s3_data_key is not None:
                        # synchronize data to the bucket before creating the stack
                        for f in find(tempd):
                            with open(f, "rb") as fd:
                                subkey = os.path.relpath(f, tempd).replace("\\", "/")
                                logging.info(
                                    "Upload %s to %s:%s%s",
                                    subkey,
                                    self.s3_bucket,
                                    self.s3_data_key,
                                    subkey,
                                )
                                s3.put_object(
                                    Bucket=self.s3_bucket,
                                    Body=fd,
                                    ServerSideEncryption="AES256",
                                    Key=self.s3_data_key + subkey,
                                )

                if self.s3_template_key is not None:
                    logging.info(
                        "Upload template to %s:%s", self.s3_bucket, self.s3_template_key
                    )
                    s3.put_object(
                        Bucket=self.s3_bucket,
                        Body=stack.body.encode("utf-8"),
                        ServerSideEncryption="AES256",
                        Key=self.s3_template_key,
                    )

                logging.info("Validate template for stack %s" % stack.name)
                try:
                    stack.validate(url=self.s3_template_url)
                except Exception:
                    logging.error("Invalid cloud formation template")
                    logging.error(stack.body)
                    raise

                if stack.exists():
                    changeset_name = "changeset%s" % int(time.time())
                    logging.info("Push changeset: %s" % changeset_name)
                    stack.create_change_set(changeset_name, url=self.s3_template_url)
                    result = stack.describe_change_set(changeset_name)
                    while result["Status"] in ("CREATE_PENDING", "CREATE_IN_PROGRESS"):
                        time.sleep(1.0)
                        result = stack.describe_change_set(changeset_name)

                    if result["Status"] == "FAILED":
                        change_executed = False
                        if (
                            "The submitted information didn't contain changes"
                            in result["StatusReason"]
                        ):
                            logging.warning(result["StatusReason"])
                            change_executed = True
                        else:
                            logging.error(result["StatusReason"])

                        stack.delete_change_set(changeset_name)
                        if not change_executed:
                            return 1
                    else:
                        for el in result["Changes"]:
                            if "ResourceChange" not in el:
                                continue
                            logging.info(
                                "%-8s %-32s: (replacement:%s)",
                                el["ResourceChange"].get("Action"),
                                el["ResourceChange"].get("LogicalResourceId"),
                                el["ResourceChange"].get("Replacement", "n/a"),
                            )

                        if self.args.apply_changeset and self._prompt_yes(
                            "Apply change"
                        ):
                            stack.execute_change_set(
                                changeset_name=changeset_name,
                                wait=self.args.wait_stack_creation,
                            )

                            if self.args.wait_stack_creation:
                                return (
                                    0
                                    if stack.state()["StackStatus"] == "UPDATE_COMPLETE"
                                    else 1
                                )
                        return 0
                else:
                    logging.info("Create new stack")
                    stack.create(
                        url=self.s3_template_url, wait=self.args.wait_stack_creation
                    )
            elif self.args.command == "show":
                print(stack.body)
            elif self.args.command == "protect":
                # Enable termination protection
                result = stack.enable_termination_protection()

                if self.stack_policy_body is not None:
                    stack.set_stack_policy(self.stack_policy_body)
                else:
                    print("No stack policy to set")
            elif self.args.command == "show-cfn-policy":
                try:
                    print(
                        json.dumps(
                            stack.cfn_policy_document().as_dict,  # type: ignore
                            indent=2,
                        )
                    )
                except AttributeError as attr_e:
                    print(f"command supported only with troposphere stacks: {attr_e}")
            elif self.args.command == "delete":
                stack.delete(wait=self.args.wait_stack_creation)

            return 0
        except botocore.exceptions.ClientError as e:
            logging.error(str(e))
            return 1

    def execute(
        self,
        args: list[str] | None = None,
        known_args_only: bool = False,
        aws_env: Session | None = None,
    ) -> int:
        """Execute application and return exit status.

        See parse_args arguments.
        """
        super(CFNMain, self).parse_args(args, known_args_only)
        assert self.args is not None

        return_val = 0
        stacks = self.create_stack()

        if isinstance(stacks, list):
            for stack in stacks:
                return_val = self.execute_for_stack(stack, aws_env=aws_env)
                # Stop at first failure
                if return_val:
                    return return_val
        else:
            return_val = self.execute_for_stack(stacks, aws_env=aws_env)

        return return_val

    @abc.abstractmethod
    def create_stack(self) -> Stack | list[Stack]:
        """Create a stack.

        :return: Stack on which the application will operate
        :rtype: Stack
        """
        pass

    @property
    def stack_policy_body(self):
        """Stack Policy that can be set by calling the command ``protect``.

        :return: the inline stack policy
        :rtype: str
        """
        return None

    def start_session(
        self,
        profile: str | None = None,
        region: str | None = None,
        aws_env: Session | None = None,
    ) -> None:
        """Start the AWS session.

        If an assume_role was passed in the constructor, then it will try
        to assume the role.

        :param profile: AWS profile for the session
        :param region: region for the session
        :param aws_env: custom AWS session to use
        """
        if aws_env is not None:
            self.aws_env = aws_env
        else:
            if self.assume_role:
                main_session = Session(regions=self.regions, profile=profile)
                self.aws_env = main_session.assume_role(
                    self.assume_role[0], self.assume_role[1]
                )
                # ??? needed since we still use a global variable for AWSEnv
                Env().aws_env = self.aws_env
            else:
                self.aws_env = AWSEnv(regions=self.regions, profile=profile)
            self.aws_env.default_region = self.regions[0] if region is None else region



from __future__ import annotations
from typing import TYPE_CHECKING
from e3.aws import name_to_id
from e3.aws.troposphere import Construct
from e3.aws.troposphere.iam.policy_document import PolicyDocument
from e3.aws.troposphere.iam.policy_statement import Allow

from troposphere import sqs, GetAtt, Ref

if TYPE_CHECKING:
    from typing import Optional
    from troposphere import AWSObject
    from e3.aws.troposphere import Stack
    from e3.aws.troposphere.iam.policy_statement import ConditionType


class Queue(Construct):
    """A SQS Topic."""

    def __init__(
        self,
        name: str,
        fifo: bool = False,
        visibility_timeout: int = 30,
        dlq_name: Optional[str] = None,
    ) -> None:
        """Initialize a SQS.

        :param name: topic name
        """
        self.name = name
        self.attr = {"QueueName": name, "VisibilityTimeout": visibility_timeout}

        if fifo:
            self.attr.update(
                {
                    "FifoQueue": True,
                    "QueueName": f"{name}.fifo",
                    "ContentBasedDeduplication": True,
                }
            )
        if dlq_name:
            self.attr["RedrivePolicy"] = {
                "deadLetterTargetArn": GetAtt(name_to_id(dlq_name), "Arn"),
                "maxReceiveCount": "3",
            }

    def allow_service_to_write(
        self, service: str, name_suffix: str, condition: Optional[ConditionType] = None
    ) -> sqs.QueuePolicy:
        """Enable a given service to send a message."""
        return sqs.QueuePolicy(
            name_to_id(f"{self.name}Policy{name_suffix}"),
            Queues=[self.ref],
            PolicyDocument=PolicyDocument(
                statements=[
                    Allow(
                        action="sqs:SendMessage",
                        resource=self.arn,
                        principal={"Service": f"{service}.amazonaws.com"},
                        condition=condition,
                    )
                ]
            ).as_dict,
        )

    @property
    def arn(self) -> GetAtt:
        """SQS ARN."""
        return GetAtt(name_to_id(self.name), "Arn")

    @property
    def ref(self) -> Ref:
        """Ref of the SQS."""
        return Ref(name_to_id(self.name))

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Compute AWS resources for the construct."""
        return [sqs.Queue.from_dict(name_to_id(self.name), self.attr)]



"""Provide events construct tests."""

from e3.aws.troposphere import Stack
from e3.aws.troposphere.ecs.cluster import FargateCluster
from e3.aws.troposphere.ecs.vpc import EcsVPC
from e3.aws.troposphere.events.rule import FargateScheduledTaskRule

EXPECTED_FARGATE_SCHEDULED_RULE = {
    "TestRule": {
        "Properties": {
            "Description": "This is a test rule",
            "Name": "test-rule",
            "ScheduleExpression": "cron(0/15 * * * ? *)",
            "State": "DISABLED",
            "Targets": [
                {
                    "Arn": {"Fn::GetAtt": ["TestCluster", "Arn"]},
                    "RoleArn": {"Fn::GetAtt": ["ECSEventsRole", "Arn"]},
                    "EcsParameters": {
                        "LaunchType": "FARGATE",
                        "NetworkConfiguration": {
                            "AwsVpcConfiguration": {
                                "AssignPublicIp": "DISABLED",
                                "SecurityGroups": [{"Ref": "TestVpcSecurityGroup"}],
                                "Subnets": [{"Ref": "TestVpcSubnet"}],
                            }
                        },
                        "TaskDefinitionArn": {"Ref": "TestTask"},
                        "PlatformVersion": "1.4.0",
                    },
                    "Id": "TestTaskTarget",
                }
            ],
        },
        "Type": "AWS::Events::Rule",
    }
}


def test_fargate_scheduled_rule(stack: Stack) -> None:
    """Test fargate scheduled rule creation."""
    stack.add(
        FargateScheduledTaskRule(
            description="This is a test rule",
            ecs_cluster=FargateCluster(name="test-cluster"),
            name="test-rule",
            schedule_expression="cron(0/15 * * * ? *)",
            task_names=["test-task"],
            vpc=EcsVPC(name="test-vpc", region="eu-west-1"),
            state="DISABLED",
        )
    )
    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == EXPECTED_FARGATE_SCHEDULED_RULE



"""Provide IAM construct tests."""

from e3.aws.troposphere.iam.role import Role
from e3.aws.troposphere.iam.policy_statement import Allow, Trust
from e3.aws.troposphere import Stack

EXPECTED_ROLE = {
    "TestRole": {
        "Properties": {
            "RoleName": "TestRole",
            "Description": "TestRole description",
            "MaxSessionDuration": 7200,
            "Path": "/",
            "AssumeRolePolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": "test"},
                        "Action": "sts:AssumeRole",
                    }
                ],
            },
            "Tags": [
                {"Key": "Name", "Value": "TestRole"},
                {"Key": "TestTagKey", "Value": "TestTagValue"},
            ],
        },
        "Type": "AWS::IAM::Role",
    }
}

EXPECTED_TRUST_ROLES = {
    "TestRole": {
        "Properties": {
            "AssumeRolePolicyDocument": {
                "Statement": [
                    {
                        "Action": ["sts:SetSourceIdentity"],
                        "Effect": "Allow",
                        "Principal": {
                            "AWS": ["arn:aws:iam::123456789012:role/OtherRole"]
                        },
                    }
                ],
                "Version": "2012-10-17",
            },
            "Description": "TestRole description",
            "Path": "/",
            "RoleName": "TestRole",
            "Tags": [{"Key": "Name", "Value": "TestRole"}],
        },
        "Type": "AWS::IAM::Role",
    },
}


def test_role(stack: Stack) -> None:
    """Test IAM role creation.

    Creating a Role also tests PolicyDocument and Policystatement classes.
    """
    stack.add(
        Role(
            name="TestRole",
            description="TestRole description",
            max_session_duration=7200,
            trust={"Service": "test"},
            tags={"TestTagKey": "TestTagValue"},
        )
    )
    assert stack.export()["Resources"] == EXPECTED_ROLE


def test_statement() -> None:
    """Test an IAM policy statement creation."""
    statement = Allow(action="s3:putObject", resource="*")
    assert statement.as_dict == {
        "Action": "s3:putObject",
        "Effect": "Allow",
        "Resource": "*",
    }


def test_trust_roles(stack: Stack) -> None:
    """Test IAM role creation.

    Creating a Role that trust another Role
    """
    stack.add(
        Role(
            name="TestRole",
            description="TestRole description",
            trust=Trust(
                roles=[(123456789012, "OtherRole")], actions=["sts:SetSourceIdentity"]
            ),
        )
    )
    assert stack.export()["Resources"] == EXPECTED_TRUST_ROLES



"""Provide ecr construct tests."""
from troposphere import ecs

from e3.aws.troposphere import Stack
from e3.aws.troposphere.ecs.cluster import FargateCluster
from e3.aws.troposphere.ecs.task_definition import FargateTaskDefinition
from e3.aws.troposphere.ecs.vpc import EcsVPC


EXPECTED_FARGATE_CLUSTER = {
    "TestCluster": {
        "Properties": {
            "ClusterName": "test-cluster",
            "ClusterSettings": [{"Name": "containerInsights", "Value": "enabled"}],
            "CapacityProviders": ["FARGATE"],
            "DefaultCapacityProviderStrategy": [
                {"CapacityProvider": "FARGATE", "Weight": "1"}
            ],
            "Tags": [{"Key": "Name", "Value": "test-cluster"}],
        },
        "Type": "AWS::ECS::Cluster",
    },
    "ECSPassExecutionRolePolicy": {
        "Properties": {
            "Description": "Needed to be attached to ECSEventsRole if scheduldedtask "
            "requires ECSTaskExecutionRole",
            "ManagedPolicyName": "ECSPassExecutionRolePolicy",
            "Path": "/",
            "PolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Action": ["iam:PassRole"],
                        "Resource": {"Fn::GetAtt": ["ECSTaskExecutionRole", "Arn"]},
                    }
                ],
            },
        },
        "Type": "AWS::IAM::ManagedPolicy",
    },
    "ECSTaskExecutionRole": {
        "Properties": {
            "RoleName": "ECSTaskExecutionRole",
            "Description": "grants the Amazon ECS container agent permission to make"
            " AWS API calls on your behalf.",
            "ManagedPolicyArns": [
                "arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy"
            ],
            "Path": "/",
            "AssumeRolePolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": ["ecs-tasks.amazonaws.com"]},
                        "Action": "sts:AssumeRole",
                    }
                ],
            },
            "Tags": [{"Key": "Name", "Value": "ECSTaskExecutionRole"}],
        },
        "Type": "AWS::IAM::Role",
    },
    "ECSEventsRole": {
        "Properties": {
            "RoleName": "ECSEventsRole",
            "Description": "Allow CloudWatch Events service to run Amazon ECS tasks",
            "ManagedPolicyArns": [
                "arn:aws:iam::aws:policy/service-role/"
                "AmazonEC2ContainerServiceEventsRole",
                {"Ref": "ECSPassExecutionRolePolicy"},
            ],
            "Path": "/",
            "AssumeRolePolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": ["events.amazonaws.com"]},
                        "Action": "sts:AssumeRole",
                    }
                ],
            },
            "Tags": [{"Key": "Name", "Value": "ECSEventsRole"}],
        },
        "Type": "AWS::IAM::Role",
    },
}


EXPECTED_ECS_VPC = {
    "TestVpc": {
        "Properties": {
            "CidrBlock": "10.0.0.0/16",
            "EnableDnsHostnames": True,
            "EnableDnsSupport": True,
            "Tags": [{"Key": "Name", "Value": "test-vpc"}],
        },
        "Type": "AWS::EC2::VPC",
    },
    "TestVpcSubnet": {
        "Properties": {"VpcId": {"Ref": "TestVpc"}, "CidrBlock": "10.0.0.0/24"},
        "Type": "AWS::EC2::Subnet",
    },
    "TestVpcSecurityGroup": {
        "Properties": {
            "GroupDescription": "Security group for ECS tasks",
            "SecurityGroupEgress": [],
            "SecurityGroupIngress": [],
            "VpcId": {"Ref": "TestVpc"},
        },
        "Type": "AWS::EC2::SecurityGroup",
    },
    "TestVpcIngress": {
        "Properties": {
            "CidrIp": "10.0.0.0/16",
            "FromPort": "443",
            "ToPort": "443",
            "IpProtocol": "tcp",
            "GroupId": {"Ref": "TestVpcSecurityGroup"},
        },
        "Type": "AWS::EC2::SecurityGroupIngress",
    },
    "TestVpcEgress": {
        "Properties": {
            "CidrIp": "10.0.0.0/16",
            "IpProtocol": "-1",
            "GroupId": {"Ref": "TestVpcSecurityGroup"},
        },
        "Type": "AWS::EC2::SecurityGroupEgress",
    },
    "TestVpcS3Egress": {
        "Properties": {
            "DestinationPrefixListId": "pl-6da54004",
            "FromPort": "443",
            "ToPort": "443",
            "IpProtocol": "tcp",
            "GroupId": {"Ref": "TestVpcSecurityGroup"},
        },
        "Type": "AWS::EC2::SecurityGroupEgress",
    },
    "TestVpcS3RouteTable": {
        "Properties": {"VpcId": {"Ref": "TestVpc"}},
        "Type": "AWS::EC2::RouteTable",
    },
    "TestVpcS3RouteTableAssoc": {
        "Properties": {
            "RouteTableId": {"Ref": "TestVpcS3RouteTable"},
            "SubnetId": {"Ref": "TestVpcSubnet"},
        },
        "Type": "AWS::EC2::SubnetRouteTableAssociation",
    },
    "TestVpcS3Endpoint": {
        "Properties": {
            "PolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": "*",
                        "Action": ["s3:GetObject", "s3:ListBucket"],
                        "Resource": "*",
                    }
                ],
            },
            "RouteTableIds": [{"Ref": "TestVpcS3RouteTable"}],
            "ServiceName": "com.amazonaws.eu-west-1.s3",
            "VpcEndpointType": "Gateway",
            "VpcId": {"Ref": "TestVpc"},
        },
        "Type": "AWS::EC2::VPCEndpoint",
    },
    "TestVpcCloudwatchLogsEndpoint": {
        "Properties": {
            "PrivateDnsEnabled": True,
            "SecurityGroupIds": [{"Ref": "TestVpcSecurityGroup"}],
            "ServiceName": "com.amazonaws.eu-west-1.logs",
            "SubnetIds": [{"Ref": "TestVpcSubnet"}],
            "VpcEndpointType": "Interface",
            "VpcId": {"Ref": "TestVpc"},
        },
        "Type": "AWS::EC2::VPCEndpoint",
    },
    "TestVpcSTSEndpoint": {
        "Properties": {
            "PrivateDnsEnabled": True,
            "SecurityGroupIds": [{"Ref": "TestVpcSecurityGroup"}],
            "ServiceName": "com.amazonaws.eu-west-1.sts",
            "SubnetIds": [{"Ref": "TestVpcSubnet"}],
            "VpcEndpointType": "Interface",
            "VpcId": {"Ref": "TestVpc"},
        },
        "Type": "AWS::EC2::VPCEndpoint",
    },
    "TestVpcEcrDkrEndpoint": {
        "Properties": {
            "PolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": "*",
                        "Action": [
                            "ecr:BatchGetImage",
                            "ecr:GetAuthorizationToken",
                            "ecr:GetDownloadUrlForLayer",
                        ],
                        "Resource": "*",
                    }
                ],
            },
            "PrivateDnsEnabled": True,
            "SecurityGroupIds": [{"Ref": "TestVpcSecurityGroup"}],
            "ServiceName": "com.amazonaws.eu-west-1.ecr.dkr",
            "SubnetIds": [{"Ref": "TestVpcSubnet"}],
            "VpcEndpointType": "Interface",
            "VpcId": {"Ref": "TestVpc"},
        },
        "Type": "AWS::EC2::VPCEndpoint",
    },
    "TestVpcEcrApiEndpoint": {
        "Properties": {
            "PolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": "*",
                        "Action": [
                            "ecr:BatchGetImage",
                            "ecr:GetAuthorizationToken",
                            "ecr:GetDownloadUrlForLayer",
                        ],
                        "Resource": "*",
                    }
                ],
            },
            "PrivateDnsEnabled": True,
            "SecurityGroupIds": [{"Ref": "TestVpcSecurityGroup"}],
            "ServiceName": "com.amazonaws.eu-west-1.ecr.api",
            "SubnetIds": [{"Ref": "TestVpcSubnet"}],
            "VpcEndpointType": "Interface",
            "VpcId": {"Ref": "TestVpc"},
        },
        "Type": "AWS::EC2::VPCEndpoint",
    },
}


EXPECTED_FARGATE_TASK_DEFINITION = {
    "TestFargateTaskDefinition": {
        "Properties": {
            "ContainerDefinitions": [
                {
                    "Image": "image-uri",
                    "Name": "container-def-name",
                    "LogConfiguration": {
                        "LogDriver": "awslogs",
                        "Options": {
                            "awslogs-group": "test-log-group",
                            "awslogs-create-group": True,
                            "awslogs-region": "eu-west-1",
                            "awslogs-stream-prefix": "test-prefix",
                        },
                    },
                }
            ],
            "Cpu": "1024",
            "Memory": "4096",
            "NetworkMode": "awsvpc",
            "RequiresCompatibilities": ["FARGATE"],
            "ExecutionRoleArn": {"Ref": "ECSTaskExecutionRole"},
            "TaskRoleArn": "task-role-name",
            "Tags": [{"Key": "Name", "Value": "test-fargate-task-definition"}],
        },
        "Type": "AWS::ECS::TaskDefinition",
    }
}


def test_fargate_cluster(stack: Stack) -> None:
    """Test Fargate cluster creation."""
    stack.add(FargateCluster(name="test-cluster"))
    assert stack.export()["Resources"] == EXPECTED_FARGATE_CLUSTER


def test_ecs_vpc(stack: Stack) -> None:
    """Test ECS VPC creation."""
    stack.add(EcsVPC(name="test-vpc", region="eu-west-1"))
    assert stack.export()["Resources"] == EXPECTED_ECS_VPC


def test_fargate_task_definition(stack: Stack) -> None:
    """Test fargate task definition creation."""
    docker_container_def = ecs.ContainerDefinition(
        Image="image-uri",
        Name="container-def-name",
        LogConfiguration=ecs.LogConfiguration(
            LogDriver="awslogs",
            Options={
                "awslogs-group": "test-log-group",
                "awslogs-create-group": True,
                "awslogs-region": "eu-west-1",
                "awslogs-stream-prefix": "test-prefix",
            },
        ),
    )

    stack.add(
        FargateTaskDefinition(
            name="test-fargate-task-definition",
            container_definitions=[docker_container_def],
            task_role_arn="task-role-name",
            cpu="1024",
            memory="4096",
        )
    )
    assert stack.export()["Resources"] == EXPECTED_FARGATE_TASK_DEFINITION



from e3.aws import session
from dateutil.parser import parse as parse_date


class EC2Element:
    """EC2 Element.

    All objects returned by EC2 API
    """

    def __init__(self, data, region):
        """Initialize an EC2 Element.

        :param data: data as returned by botocore
        :type data: dict
        :param region: region of the EC2 object
        :type region: str
        """
        self.data = data
        self.region = region

        # Compute tags
        if "Tags" in self.data:
            self.tags = {el["Key"]: el["Value"] for el in self.data["Tags"]}
        else:
            self.tags = {}

        # Map botocore attributes declared in PROPERTIES into Python
        # properties.
        for key, name in self.PROPERTIES.items():
            setattr(
                self.__class__,
                name,
                property(lambda s, default_key=key: s.data.get(default_key)),
            )

    @property
    def logical_id(self):
        """Cloud Formation logical id.

        :return: the id or None (if the element is not part of a stack)
        :rtype: str | None
        """
        return self.tags.get("aws:cloudformation:logical-id")


class SecurityGroup(EC2Element):
    """Security Group description."""

    PROPERTIES = {"GroupName": "group_name", "GroupId": "group_id"}

    @session()
    def __init__(self, group_id=None, region=None, data=None, session=None):
        """Initialize a security group.

        :param group_id: group id. Used to retrieve group data
            if data parameter is None
        :type group_id: str | None
        :param region: region of the security group
        :type region: str | None
        :param data: botocore data describing the group. If None
            then data is fetched automatically using group id and
            region
        """
        if data is None:
            assert region is not None and group_id is not None
            data = session.client("ec2", region).describe_security_groups(
                GroupIds=[group_id]
            )["SecurityGroups"][0]
        super().__init__(data, region)

    @classmethod
    @session()
    def ls(cls, filters=None, session=None):
        """List user security groups.

        Note that this API is cross region. All regions used when creating
        the session are used.

        :param filters: same as Filters parameters of describe_security_groups
            (see botocore)
        :type filters: dict
        :return a list of images
        :rtype: list[SecurityGroup]
        """
        if filters is None:
            filters = []
        result = []
        for r in session.regions:
            c = session.client("ec2", r)
            region_result = c.describe_security_groups(Filters=filters)
            for sg in region_result["SecurityGroups"]:
                result.append(SecurityGroup(sg["GroupId"], r, data=sg))
        return result


class Instance(EC2Element):
    """EC2 Instance."""

    PROPERTIES = {"InstanceId": "instance_id"}

    @session()
    def __init__(self, instance_id=None, region=None, data=None, session=None):
        """Initialize an EC2 instance description.

        :param instance_id: instance id. Used to retrieve instance data
            if data parameter is None
        :type group_id: str | None
        :param region: region of the security group
        :type region: str | None
        :param data: botocore data describing the instance. If None
            then data is fetched automatically using instance id and
            region
        """
        if data is None:
            assert instance_id is not None and region is not None
            data = session.client("ec2", region).describe_instances(
                InstanceIds=[instance_id]
            )["Reservations"]["Instances"][0]

        super().__init__(data, region)

    @property
    def security_groups(self):
        """List security groups attached to the instance.

        :return: a list of security groups
        :rtype: list[SecurityGroup]
        """
        return [
            SecurityGroup(el["GroupId"], region=self.region)
            for el in self.data["SecurityGroups"]
        ]

    @property
    def network_interfaces(self):
        """List network interfaces attached to an instance.

        :return: a list of network interface
        :rtype: list[NetworkInterface]
        """
        return [
            NetworkInterface(ni, region=self.region)
            for ni in self.data.get("NetworkInterfaces", [])
        ]

    @property
    def has_public_ip(self):
        """Check if instance has a public IP.

        :return: True if at least one interface has a public IP
        :rtype: bool
        """
        for ni in self.network_interfaces:
            if ni.public_ip is not None:
                return True
        return False

    @property
    def block_device_mappings(self):
        """List block device mappings.

        :return: the list of mappings
        :rtype: list[BlockDeviceMapping]
        """
        return [
            BlockDeviceMapping(bdm, region=self.region)
            for bdm in self.data["BlockDeviceMappings"]
        ]

    @classmethod
    @session()
    def ls(cls, filters=None, session=None):
        """List user instances.

        Note that instances in "terminated" mode are ignored.

        :param filters: same as Filters parameters of describe_instances
            (see botocore)
        :type filters: dict
        :return a list of instances
        :rtype: list[Instance]
        """
        if filters is None:
            filters = []
        result = []
        for r in session.regions:
            c = session.client("ec2", r)
            region_result = c.describe_instances(Filters=filters)
            for reservation in region_result["Reservations"]:
                if "Instances" not in reservation:
                    continue
                for instance_data in reservation["Instances"]:
                    if instance_data["State"]["Name"] == "terminated":
                        # Ignore instance that are going to disappear soon
                        continue
                    result.append(
                        Instance(
                            instance_data["InstanceId"],
                            r,
                            data=instance_data,
                            session=session,
                        )
                    )
        return result


class BlockDeviceMapping(EC2Element):
    """Block Device Mappping."""

    PROPERTIES = {"DeviceName": "device_name"}

    @property
    def is_ebs(self):
        """Check if mapping is an EBS.

        :return: True if this is an EBS, False otherwise
        :rtype: bool
        """
        return "Ebs" in self.data

    @property
    def encrypted(self):
        """Check if the block device is encrypted.

        :return: True if encrypted, false otherwise
        :rtype: bool
        """
        return "Ebs" in self.data and self.data["Ebs"].get("Encrypted", False)

    @property
    def snapshot_id(self):
        """Retrieve snapshot id.

        :return: the associated snaptshot if it exist
        :rtype: str | None
        """
        if self.is_ebs:
            return self.data["Ebs"].get("SnapshotId")
        else:
            return None


class VolumeAttachment(EC2Element):
    """Volume Attachment."""

    PROPERTIES = {
        "InstanceId": "instance_id",
        "Device": "device",
        "State": "state",
        "VolumeId": "volume_id",
        "DeleteOnTermination": "delete_on_termination",
    }

    def __init__(self, data, region):
        """Initialize Volume Attachment description."""
        super().__init__(data, region)
        self._instance_cache = None
        self._volume_cache = None

    @property
    def attach_time(self):
        """Return time at which attachment was done.

        :return: attach time
        :rtype: datetime.datetime
        """
        return parse_date(self.data["AttachTime"])

    @property
    def instance(self):
        """Retrieve instance attached to the volume.

        :return: an Instance object
        :rtype: Instance
        """
        if self._instance_cache is None:
            self._instance_cache = Instance(self.instance_id, region=self.region)
        return self._instance_cache

    @property
    def volume(self):
        """Retrieve the Volume.

        :return: the Volume part of the attachment.
        :rtype: Volume
        """
        if self._volume_cache is None:
            self._volume_cache = Volume(self.volume_id, region=self.region)
        return self._volume_cache


class Volume(EC2Element):
    """EC2 Volume."""

    PROPERTIES = {
        "VolumeId": "volume_id",
        "AvailabilityZone": "availability_zone",
        "Size": "size",
        "SnapshotId": "snapshot_id",
        "State": "state",
        "VolumeType": "volume_type",
        "Encrypted": "encrypted",
    }

    @session()
    def __init__(self, volume_id=None, region=None, data=None, session=None):
        """Initialize an EC2 volume description.

        :param volume_id: volume id. Used to retrieve volume data
            if data parameter is None
        :type group_id: str | None
        :param region: region of the volume
        :type region: str | None
        :param data: botocore data describing the volume. If None
            then data is fetched automatically using volume id and
            region
        """
        if data is None:
            assert volume_id is not None and region is not None
            data = session.client("ec2", region).describe_volumes(
                InstanceIds=[volume_id]
            )["Volumes"][0]

        super().__init__(data, region)

    @property
    def attachments(self):
        """List attachments that involve that volume.

        :return: a list of attachments
        :rtype: list[VolumeAttachment]
        """
        return [VolumeAttachment(va, self.region) for va in self.data["Attachments"]]

    @property
    def create_time(self):
        """Return creation time.

        :return: time at which volume was created
        :rtype: datetime.datetime
        """
        return parse_date(self.data["CreateTime"])

    @session()
    def delete(self):
        """Delete a volume."""
        session.client("ec2", self.region).delete_volume(VolumeId=self.volume_id)

    @classmethod
    @session()
    def ls(cls, filters=None):
        """List user AMIs.

        :param filters: same as Filters parameters of describe_volumes
            (see botocore)
        :type filters: dict
        :return a list of volumes
        :rtype: list[Volume]
        """
        if filters is None:
            filters = []
        result = []
        for r in session.regions:
            c = session.client("ec2", r)
            region_result = c.describe_volumes(Filters=filters)
            for volume in region_result.get("Volumes", []):
                result.append(Volume(volume["VolumeId"], r, data=volume))
        return result


class Snapshot(EC2Element):
    PROPERTIES = {"Encrypted": "encrypted", "SnapshotId": "snapshot_id"}

    @session()
    def __init__(self, snapshot_id=None, region=None, data=None, session=None):
        """Initialize an EC2 snapshot description.

        :param snapshot_id: snapshot id. Used to retrieve snapshot data
            if data parameter is None
        :type snapshot_id: str | None
        :param region: region of the snapshot
        :type region: str | None
        :param data: botocore data describing the snapshot. If None
            then data is fetched automatically using snapshot id and
            region
        """
        if data is None:
            assert snapshot_id is not None and region is not None
            data = session.client("ec2", region).describe_snapshots(
                SnapshotIds=[snapshot_id]
            )["Snapshots"][0]

        super().__init__(data, region)

    @session()
    def delete(self, session=None):
        """Delete a snapshot."""
        session.client("ec2", self.region).delete_snapshot(SnapshotId=self.snapshot_id)

    @classmethod
    @session()
    def ls(cls, filters=None, session=None):
        """List snapshots.

        :param filters: same as Filters parameters of describe_snapshots
            (see botocore)
        :type filters: dict
        :return a list of snaptshots
        :rtype: list[Snapshot]
        """
        if filters is None:
            filters = []
        result = []
        for r in session.regions:
            c = session.client("ec2", r)
            region_result = c.describe_snapshots(OwnerIds=["self"], Filters=filters)
            for snapshot in region_result.get("Snapshots", []):
                result.append(Snapshot(snapshot["SnapshotId"], r, data=snapshot))
        return result


class NetworkInterface(EC2Element):
    """EC2 Network Interface."""

    def public_ip(self):
        """Return public ip.

        :return: return public ip or None
        :rtype: str | None
        """
        if "Association" in self.data:
            return self.data["Association"].get("PublicIp")
        else:
            return None



"""Provide fixtures for e3 aws troposphere tests."""

import pytest

from e3.aws.troposphere import Stack


@pytest.fixture
def stack() -> Stack:
    """Stack fixture to help dumping dictionnaries from constructs."""
    return Stack("test-stack", "this is a test stack")
